
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <cblas.h>

#ifdef CUDA
    #include <cuda.h>
    #include <cublas_v2.h>
#endif

///////////////////////////////////////////////////////////////////////////////
/**
 * @brief Get time.
 * Return current time in seconds as a double-precision floating-point number.
 */
double pulsar_time_of_day()
{
    struct timeval  time_val;
    struct timezone time_zone;
    gettimeofday(&time_val, &time_zone);
    return (double)(time_val.tv_sec) + (double)(time_val.tv_usec) / 1000000.0;
}

///////////////////////////////////////////////////////////////////////////////
/**
 * @brief Warm up BLAS by calling DGEMM.
 */
void pulsar_blas_warmup()
{
    double *a = (double*)malloc(256*256*sizeof(double));
    double *b = (double*)malloc(256*256*sizeof(double));
    double *c = (double*)malloc(256*256*sizeof(double));
    cblas_dgemm(
        CblasColMajor,
        CblasNoTrans, CblasNoTrans,
        256, 256, 256,
        1.0, a, 256,
             b, 256,
        1.0, c, 256);
    free(a);
    free(b);
    free(c);
}

///////////////////////////////////////////////////////////////////////////////
/**
 * @brief Warm up BLAS by calling DGEMM.
 */
void pulsar_cublas_warmup()
{
#ifdef CUDA
    double *d_A;
    double *d_B;
    double *d_C;
    cudaMalloc((void**)&d_A, 256*256*sizeof(double));
    cudaMalloc((void**)&d_B, 256*256*sizeof(double));
    cudaMalloc((void**)&d_C, 256*256*sizeof(double));
    cublasHandle_t handle;
    cublasCreate(&handle);
    double alpha = 1.0;
    double beta = 1.0;
    cublasDgemm(handle,
        CUBLAS_OP_N, CUBLAS_OP_N,
        256, 256, 256,
        &alpha, d_A, 256,
                d_B, 256,
        &beta,  d_C, 256);
    cublasDestroy(handle);
    cudaFree(d_A);
    cudaFree(d_B);
    cudaFree(d_C);
#endif
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Initialize a tile with random numbers.
 * Call DLARNV with the seed derived from the tile's position.
 *
 * @param tile
 * @param m vertical position
 * @param n horizontal position
 * @param nb tile size
 */
void pulsar_tile_init(double *tile, int m, int n, int nb)
{
    int iseed[4];
    iseed[0] = m & 0x0FFF;
    iseed[1] = n & 0x0FFF;
    iseed[2] = ((m >> 12) + (n >> 12)) & 0x0FFF;
    iseed[3] = 1;
    LAPACKE_dlarnv_work(1, iseed, nb*nb, tile);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Print corners of a tile.
 * Print 2x2 corners of a tile.
 *
 * @param tile
 * @param nb tile size
 */
void pulsar_tile_print(double *tile, int nb)
{
    int i, j;
    printf("\n");
    for (j = 0; j < nb; j++) {
        if (j == 2) {
            j = nb-2;
            printf("  ...   ...   ...   ...   ... \n");
        }
        for (i = 0; i < nb; i++) {
            if (i == 2) {
                i = nb-2;
                printf("  ...");
            }
            printf("%6.2lf", tile[i*nb+j]);
        }
        printf("\n");
    }
}
