
#include <prt.h>
#include <cblas.h>
#include <plasma.h>

#ifdef CUDA
    #include <cuda.h>
    #include <cublas_v2.h>
#endif

#include "pulsar_util.h"

/**
 * @brief tile QR VSA's global store
 */
typedef struct {
    double **A;
    int ib;
    int nb;
    int mt;
    int nt;
} qr_global_store_t;

/**
 * @brief tile QR VDP's local store
 */
typedef struct {
    prt_packet_t *A1;
    prt_packet_t *A2;
    prt_packet_t *T;
    prt_packet_t *V;
    prt_packet_t *work;
    prt_packet_t *tau;
#ifdef CUDA
    cublasHandle_t cublas_handle;
#endif
} qr_local_store_t;

prt_mapping_t qr_vdp_mapping(
    int *tuple, void *global_store, int num_cores, int num_accelerators);
void qr_vsa_init(prt_vsa_t *vsa, int nb, int ib, int mt, int nt);
void qr_vdp_function1(prt_vdp_t *vdp);
void qr_vdp_function2(prt_vdp_t *vdp);
double plasma_tile_qr(int nb, int ib, int mt, int nt, int num_threads);

#ifdef CUDA
void CORE_dormqr_acc(
    PLASMA_enum side, PLASMA_enum trans,
    int m, int n, int k, int ib,
    const double *d_V, int ldv,
    const double *d_T, int ldt,
    double *d_A, int lda,
    double *d_work, int ldwork,
    cublasHandle_t cublas_handle);
#endif

#ifdef CUDA
void CORE_dtsmqr_acc(
    PLASMA_enum side, PLASMA_enum trans,
    int m1, int n1, int m2, int n2, int k, int ib,
    double *d_A1, int lda1,
    double *d_A2, int lda2,
    const double *d_V, int ldv,
    const double *d_T, int ldt,
    double *d_work, int ldwork,
    cublasHandle_t cublas_handle);
#endif

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief main
 * Set up and run the VSA for the tile QR factorization (multi-node, multi-core).
 * Optionally call PLASMA tile QR factorization (single-node, multi-core).
 * Print corner tiles (bottom-right) for comparison.
 * Measure and report performance for each case.
 *
 * @param argc
 * @param argv
 *
 * @retval 0 success
 * @retval EXIT_FAILURE error
 */
int main(int argc, char **argv)
{
    if (argc < 6) {
        printf("usage: mpirun -n <proc> ./tile_qr "
               "<ib> <nb> <tiles> <threads> [test] [trace] [lazy|aggressive]\n"
               "  proc       number of MPI processes\n"
               "  ib         inner blocking, e.g., 32, 64, ...\n"
               "  nb         tile size, e.g., 192, 256, ...\n"
               "  tiles      number of tiles in each row and each column\n"
               "  threads    number of threads per process\n"
               "             (use cores-1 if using MPI)\n"
               "  devices    number of devices (accelerators) per process\n"
               "  test       compare against PLASMA on one node\n\n"
               "  trace      print trace to an SVG file\n"
               "  lazy       lazy scheduling (breadth-first)\n"
               "  aggressive aggressive scheduling (depth-first)\n\n");
        exit(EXIT_FAILURE);
    }
    int ib = atoi(argv[1]);
    int nb = atoi(argv[2]);
    int bb = atoi(argv[3]);
    int num_threads = atoi(argv[4]);
    int num_devices = atoi(argv[5]);
    int mt = bb;
    int nt = bb;

    int arg;
    int test = 0;
    int trace = 0;
    prt_config_value_t scheduling = PRT_VDP_SCHEDULING_LAZY;
    for (arg = 5; arg < argc; arg++) {
        if (strcmp(argv[arg], "test") == 0)
            test = 1;
        else if (strcmp(argv[arg], "trace") == 0)
            trace = 1;
        else if (strcmp(argv[arg], "aggressive") == 0)
            scheduling = PRT_VDP_SCHEDULING_AGGRESSIVE;
    }
    qr_global_store_t gs;
    gs.ib = ib;
    gs.nb = nb;
    gs.mt = mt;
    gs.nt = nt;

    int mpi_rank = 0;
    int mpi_size = 1;
#ifdef MPI
    assert(MPI_Init(&argc, &argv) == MPI_SUCCESS);
    assert(MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank) == MPI_SUCCESS);
    assert(MPI_Comm_size(MPI_COMM_WORLD, &mpi_size) == MPI_SUCCESS);
#endif

    int m;
    int n;
    // Allocate the matrix where appropriate.
    gs.A = (double**)malloc(mt*nt*sizeof(double*));
    for (m = 0; m < mt; m++) {
        for (n = 0; n < nt; n++) {
            int num_cores = mpi_size * num_threads;
            int num_accelerators = mpi_size * num_devices;
            prt_mapping_t mapping =
                qr_vdp_mapping(
                    prt_tuple_new2(0, n), &gs, num_cores, num_accelerators);
            int node;
            if (mapping.location == PRT_LOCATION_HOST)
                node = mapping.rank / num_threads;
            else
                node = mapping.rank / num_devices;
            if (node == mpi_rank) {
                gs.A[n*mt+m] = (double*)malloc(nb*nb*sizeof(double));
                pulsar_tile_init(gs.A[n*mt+m], m, n, nb);
            }
        }
    }
    prt_vsa_t *vsa = prt_vsa_new(
        num_threads, num_devices, &gs, qr_vdp_mapping);
    prt_vsa_config_set(vsa, PRT_VDP_SCHEDULING, scheduling);
    prt_vsa_thread_warmup_func_set(vsa, pulsar_blas_warmup);
    prt_vsa_device_warmup_func_set(vsa, pulsar_cublas_warmup);
    if (trace)
        prt_vsa_config_set(vsa, PRT_SVG_TRACING, PRT_SVG_TRACING_ON);

    double time;
    qr_vsa_init(vsa, nb, ib, mt, nt);
    time = prt_vsa_run(vsa);
    prt_vsa_delete(vsa);

    if (mpi_rank == 0) {
        double gops = 4.0/3.0*nb*nb*nb*bb*bb*bb/1000000000.0;
        printf("\nPULSAR\n");
        printf("  nodes:   %d\n", mpi_size);
        printf("  cores:   %d\n", mpi_size*num_threads);
        printf("  devices: %d\n", mpi_size*num_devices);
        printf("  seconds: %lf\n", time);
        printf("  gflops:  %lf\n", gops/time);

        if (test) {
            time = plasma_tile_qr(nb, ib, mt, nt, num_threads);
            printf("\nPLASMA\n");
            printf("  nodes:   %d\n", 1);
            printf("  cores:   %d\n", num_threads);
            printf("  seconds: %lf\n", time);
            printf("  gflops:  %lf\n", gops/time);
        }
        printf("\n");
    }
#ifdef MPI
    assert(MPI_Finalize() == MPI_SUCCESS);
#endif
    return 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Defines z-curve mapping of VDPs to cores.
 *
 * 0 1 2 3 0 1
 *   2 3 0 1 2
 *     3 0 1 2
 *       3 0 1
 *         2 3
 *           0
 *
 * Assigns diagonal cores to CPUs and off-diagonal cores to GPUs.
 *
 * @param tuple VDP's tuple
 * @param global_store VSA's global store
 * @param num_cores VSA's (total) number of cores
 *
 * @return VDP's core (global) number
 */
prt_mapping_t qr_vdp_mapping(
    int *tuple, void *global_store, int num_cores, int num_accelerators)
{
    qr_global_store_t *gs = (qr_global_store_t*)global_store;
    int mt = gs->mt;
    int nt = gs->nt;
    int m = tuple[0];
    int n = tuple[1];

    prt_mapping_t mapping;

#ifdef CUDA
    if (m == n) {
        mapping.location = PRT_LOCATION_HOST;
        mapping.rank = (mt*m+n - m*(m+1)/2) % num_cores;
    }
    else {
        mapping.location = PRT_LOCATION_DEVICE;
        mapping.rank = (mt*m+n - m*(m+1)/2) % num_accelerators;        
    }
#else
    mapping.location = PRT_LOCATION_HOST;
    mapping.rank = (mt*m+n - m*(m+1)/2) % num_cores;
#endif

    return (mapping);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief VSA construction
 * Create the VDPs.
 * Create the channels.
 * Insert the channels in the VDPs.
 * Insert the VDPs in the VSA.
 *
 * @param vsa
 @ @param nb matrix tile size
 * @param ib QR inner blocking
 * @param mt matrix height (number of tiles)
 * @param nt matrix width (number of tiles)
 */
void qr_vsa_init(prt_vsa_t *vsa, int nb, int ib, int mt, int nt)
{
    int m;
    int n;
    prt_channel_t *channel;

    for (m = 0; m < mt; m++) {
        for (n = m; n < nt; n++) {
            prt_vdp_t *vdp; // Create the VDP.
            if (m == n)
                vdp = prt_vdp_new(
                    prt_tuple_new2(m, n), nt-m, qr_vdp_function1,
                    sizeof(qr_local_store_t), 3, 3, SteelBlue);
            else
                vdp = prt_vdp_new(
                    prt_tuple_new2(m, n), nt-m, qr_vdp_function2,
                    sizeof(qr_local_store_t), 3, 3, MediumSeaGreen);

            if (m > 0) {    // input channel 0
                channel = prt_channel_new(nb*nb*sizeof(double),
                    prt_tuple_new2(m-1, n), 0, prt_tuple_new2(m, n), 0);
                prt_vdp_channel_insert(vdp, channel, PRT_INPUT_CHANNEL, 0);
            }
            if (n > m) {    // input channel 1
                channel = prt_channel_new(nb*nb*sizeof(double),
                    prt_tuple_new2(m, n-1), 1, prt_tuple_new2(m, n), 1);
                prt_vdp_channel_insert(vdp, channel, PRT_INPUT_CHANNEL, 1);
            }
            if (n > m) {    // input channel 2
                channel = prt_channel_new(ib*nb*sizeof(double),
                    prt_tuple_new2(m, n-1), 2, prt_tuple_new2(m, n), 2);
                prt_vdp_channel_insert(vdp, channel, PRT_INPUT_CHANNEL, 2);
            }
            if (m < n) {    // output channel 0
                channel = prt_channel_new(nb*nb*sizeof(double),
                    prt_tuple_new2(m, n), 0, prt_tuple_new2(m+1, n), 0);
                prt_vdp_channel_insert(vdp, channel, PRT_OUTPUT_CHANNEL, 0);
            }
            if (n < nt-1) { // output channel 1
                channel = prt_channel_new(nb*nb*sizeof(double),
                    prt_tuple_new2(m, n), 1, prt_tuple_new2(m, n+1), 1);
                prt_vdp_channel_insert(vdp, channel, PRT_OUTPUT_CHANNEL, 1);
            }
            if (n < nt-1) { // output channel 2
                channel = prt_channel_new(ib*nb*sizeof(double),
                    prt_tuple_new2(m, n), 2, prt_tuple_new2(m, n+1), 2);
                prt_vdp_channel_insert(vdp, channel, PRT_OUTPUT_CHANNEL, 2);
            }
            // Insert the VDP.
            prt_vsa_vdp_insert(vsa, vdp);
      }
    }
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Defines operation for an edge VDP.
 * A(input 0)
 *     |
 *    _V_
 *   |   |--> V(output 1)
 *   |___|--> T(output 2)
 *
 * @param vdp
 */
void qr_vdp_function1(prt_vdp_t *vdp)
{
    qr_global_store_t *gs = (qr_global_store_t*)vdp->global_store;
    qr_local_store_t *ls = (qr_local_store_t*)vdp->local_store;

    int ib = gs->ib;
    int nb = gs->nb;
    int nt = gs->nt;
    int mt = gs->mt;
    int m = vdp->tuple[0];
    int n = vdp->tuple[1];
    int nb_size = nb*sizeof(double);
    int nbnb_size = nb*nb*sizeof(double);
    int ibnb_size = ib*nb*sizeof(double);

    // IF the first step.
    if (vdp->counter == nt-m) {
        // Allocate workspace.
        ls->work = prt_vdp_packet_new(vdp, ibnb_size, NULL);
        ls->tau  = prt_vdp_packet_new(vdp, nb_size, NULL);
        // IF the first row.
        if (m == 0) {
            // Create A1.
            void *tile = gs->A[n*mt+(nt-vdp->counter)];
            ls->A1 = prt_vdp_packet_new(vdp, nbnb_size, tile);
        }
        // Lower rows.
        else {
            // Pop A1 from a channel.
            ls->A1 = prt_vdp_channel_pop(vdp, 0);
        }
        ls->T = prt_vdp_packet_new(vdp, ibnb_size, NULL);
        memset(ls->T->data, 0, ibnb_size);
        CORE_dgeqrt(
            nb, nb, ib,
            ls->A1->data, nb,
            ls->T->data, ib,
            ls->tau->data,
            ls->work->data);
        if (n < nt-1) {
            prt_vdp_channel_push(vdp, 1, ls->A1);
            prt_vdp_channel_push(vdp, 2, ls->T);
        }
        prt_vdp_packet_release(vdp, ls->T);
    }
    // Follow up steps.
    else {
        // IF the first row.
        if (m == 0) {
            // Create A2.
            void *tile = gs->A[n*mt+(nt-vdp->counter)];
            ls->A2 = prt_vdp_packet_new(vdp, nbnb_size, tile);
        }
        // Lower rows.
        else {
            // Pop A2 from a channel.
            ls->A2 = prt_vdp_channel_pop(vdp, 0);
        }
        ls->T = prt_vdp_packet_new(vdp, ibnb_size, NULL);
        memset(ls->T->data, 0, ibnb_size);
        CORE_dtsqrt(
            nb, nb, ib,
            ls->A1->data, nb,
            ls->A2->data, nb,
            ls->T->data, ib,
            ls->tau->data,
            ls->work->data);
        prt_vdp_channel_push(vdp, 1, ls->A2);
        prt_vdp_channel_push(vdp, 2, ls->T);
        prt_vdp_packet_release(vdp, ls->A2);
        prt_vdp_packet_release(vdp, ls->T);
    }
    // IF the last step.
    if (vdp->counter == 1) {
        // IF the corner VDP.
        if (m == mt-1 && n == nt-1) {
            // Print the corner tile.
            pulsar_tile_print((double*)ls->A1->data, nb);
        }
        // Release A1 and the workspace.
        prt_vdp_packet_release(vdp, ls->A1);
        prt_vdp_packet_release(vdp, ls->work);
        prt_vdp_packet_release(vdp, ls->tau);
    }
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Defines operation for an out-of-edge VDP.
 *             A(input 0)
 *                 |
 *                _V_
 * V(input 1) -->|   |--> V(output 1) - pass-through
 * T(input 2) -->|___|--> T(output 2) - pass-through
 *                 |
 *                 V
 *            A(output 0)
 *
 * @param vdp
 */
void qr_vdp_function2(prt_vdp_t *vdp)
{
    qr_global_store_t *gs = (qr_global_store_t*)vdp->global_store;
    qr_local_store_t *ls = (qr_local_store_t*)vdp->local_store;

    int ib = gs->ib;
    int nb = gs->nb;
    int nt = gs->nt;
    int mt = gs->mt;
    int m = vdp->tuple[0];
    int n = vdp->tuple[1];
    int nbnb_size = nb*nb*sizeof(double);
    int ibnb_size = ib*nb*sizeof(double);

#ifdef CUDA
    // IF first step.
    if (vdp->counter == nt-m) {
        cublasStatus_t cublas_status;
        // Initialize VDP's CUBLAS handle.
        cublas_status = cublasCreate(&ls->cublas_handle);
        assert(cublas_status == CUBLAS_STATUS_SUCCESS);
        // Assign the VDP's stream to the VDP's CUBLAS handle.
        cublas_status = cublasSetStream(ls->cublas_handle, vdp->stream);
        prt_assert(cublas_status == CUBLAS_STATUS_SUCCESS, "cublasSetStream failed");
    }
#endif
    // IF the first step.
    if (vdp->counter == nt-m) {
        // Allocate workspace.
        ls->work = prt_vdp_packet_new(vdp, 2*ibnb_size, NULL);
        // IF the first row.
        if (m == 0) {       
            // Create A1.                                    
            void *tile = gs->A[n*mt+(nt-vdp->counter)];
#ifdef CUDA
            ls->A1 = prt_vdp_packet_new_host_to_device(
                vdp, nbnb_size, tile);
#else
            ls->A1 = prt_vdp_packet_new(vdp, nbnb_size, tile);
#endif            
        }
        // Lower rows.
        else {
            // Pop A1 from a channel.
            ls->A1 = prt_vdp_channel_pop(vdp, 0);
        }
        ls->V = prt_vdp_channel_pop(vdp, 1);
        ls->T = prt_vdp_channel_pop(vdp, 2);
        if (n < nt-1) {
            prt_vdp_channel_push(vdp, 1, ls->V);
            prt_vdp_channel_push(vdp, 2, ls->T);
        }
        // IF host VDP.
        if (vdp->location == PRT_LOCATION_HOST) {
            CORE_dormqr(
                 PlasmaLeft, PlasmaTrans,
                 nb, nb, nb, ib,
                 ls->V->data, nb,
                 ls->T->data, ib,
                 ls->A1->data, nb,
                 ls->work->data, nb);
        }
#ifdef CUDA
        // Device VDP.
        else {
            CORE_dormqr_acc(
                PlasmaLeft, PlasmaTrans,
                nb, nb, nb, ib,
                ls->V->data, nb,
                ls->T->data, ib,
                ls->A1->data, nb,
                ls->work->data, ib,
                ls->cublas_handle);
        }
#endif
        prt_vdp_packet_release(vdp, ls->V);
        prt_vdp_packet_release(vdp, ls->T);
    }
    // Follow up steps.
    else {
        // IF first row.
        if (m == 0) {
            // Create A2.
            void *tile = gs->A[n*mt+(nt-vdp->counter)];
#ifdef CUDA
            ls->A2 = prt_vdp_packet_new_host_to_device(
                vdp, nbnb_size, tile);
#else
            ls->A2 = prt_vdp_packet_new(vdp, nbnb_size, tile);
#endif
        }
        else {
            // Pop A2 from a channel.
            ls->A2 = prt_vdp_channel_pop(vdp, 0);
        }
        ls->V = prt_vdp_channel_pop(vdp, 1);
        ls->T = prt_vdp_channel_pop(vdp, 2);
        if (n < nt-1) {
            prt_vdp_channel_push(vdp, 1, ls->V);
            prt_vdp_channel_push(vdp, 2, ls->T);
        }
        // IF host VDP.
        if (vdp->location == PRT_LOCATION_HOST) {
            CORE_dtsmqr(
                 PlasmaLeft, PlasmaTrans,
                 nb, nb, nb, nb, nb, ib,
                 ls->A1->data, nb,
                 ls->A2->data, nb,
                 ls->V->data, nb,
                 ls->T->data, ib,
                 ls->work->data, ib);
        }
#ifdef CUDA
        // Device VDP.
        else {
            CORE_dtsmqr_acc(
                PlasmaLeft, PlasmaTrans,
                nb, nb, nb, nb, nb, ib,
                ls->A1->data, nb,
                ls->A2->data, nb,
                ls->V->data, nb,
                ls->T->data, ib,
                ls->work->data, ib,
                ls->cublas_handle);
        }
#endif
        prt_vdp_channel_push(vdp, 0, ls->A2);
        prt_vdp_packet_release(vdp, ls->V);
        prt_vdp_packet_release(vdp, ls->T);
        prt_vdp_packet_release(vdp, ls->A2);
    }
    // IF the last step.
    if (vdp->counter == 1) {
        // Release A1 and the workspace.
        prt_vdp_packet_release(vdp, ls->A1);
        prt_vdp_packet_release(vdp, ls->work);
    }
#ifdef CUDA
    // IF the last step.
    if (vdp->counter == 1) {
        cublasStatus_t cublas_status;
        // Destroy VDP's CUBLAS handle.
        cublas_status = cublasDestroy(ls->cublas_handle);
        assert(cublas_status == CUBLAS_STATUS_SUCCESS);
    }
#endif
}

////////////////////////////////////////////////////////////////////////////////
#ifdef CUDA
void CORE_dormqr_acc(
    PLASMA_enum side, PLASMA_enum trans,
    int m, int n, int k, int ib,
    const double *d_V, int ldv,
    const double *d_T, int ldt,
    double *d_A, int lda,
    double *d_work, int ldwork,
    cublasHandle_t cublas_handle)
{
    int nb = m;
    double *d_work2 = d_work + ib*nb;

    int j;
    for (j = 0; j < nb; j+=ib)
    {
        // W <= A1
        double done = 1.0;
        double dzero = 0.0;
        cublasDgeam(cublas_handle,
            CUBLAS_OP_N, CUBLAS_OP_N,
            ib, nb,
            &done,  &d_A[j] , lda,
            &dzero,  d_work2, ldwork,
                     d_work , ldwork);

        // W2 = V1' W
        cublasDtrmm(cublas_handle,
            CUBLAS_SIDE_LEFT,
            CUBLAS_FILL_MODE_LOWER,
            CUBLAS_OP_T, CUBLAS_DIAG_UNIT,
            ib, nb,
            &done, &d_V[j*ldv+j], ldv,
                    d_work      , ldwork,
                    d_work2     , ldwork);

        // W2 = W2 + V2' A2
      if (j < nb-ib) {
        cublasDgemm(cublas_handle,
            CUBLAS_OP_T, CUBLAS_OP_N,
            ib, nb, nb-ib-j,
            &done, &d_V[j*ldv+ib+j], ldv,
                   &d_A[      ib+j], lda,
            &done, d_work2         , ldwork);
      }

        // W = T' W2
        cublasDgemm(cublas_handle,
            CUBLAS_OP_T, CUBLAS_OP_N,
            ib, nb, ib,
            &done, &d_T[j*ldt], ldt,
                    d_work2   , ldwork,
            &dzero, d_work    , ldwork);

        // A2 = A2 - V2 W
      if (j < nb-ib) {
        double dmone = -1.0;
        cublasDgemm(cublas_handle,
            CUBLAS_OP_N, CUBLAS_OP_N,
            nb-ib-j, nb, ib,
            &dmone, &d_V[j*ldv+ib+j], ldv,
                     d_work         , ldwork,
            &done,  &d_A[      ib+j], lda);
      }

        // W2 = V1 W
        cublasDtrmm(cublas_handle,
            CUBLAS_SIDE_LEFT,
            CUBLAS_FILL_MODE_LOWER,
            CUBLAS_OP_N, CUBLAS_DIAG_UNIT,
            ib, nb,
            &done, &d_V[j*ldv+j], ldv,
                    d_work      , ldwork,
                    d_work2     , ldwork);

        // A1 = A1 - W2
        // W <- A1
        cublasDgeam(cublas_handle,
            CUBLAS_OP_N, CUBLAS_OP_N,
            ib, nb,
            &done,  &d_A[j] , lda,
            &dzero,  d_work2, ldwork,
                     d_work , ldwork);
        // A1 = W - W2
        double dmone = -1.0;
        cublasDgeam(cublas_handle,
            CUBLAS_OP_N, CUBLAS_OP_N,
            ib, nb,
            &done,  d_work , ldwork,
            &dmone, d_work2, ldwork,
                    &d_A[j], lda);
    }
}
#endif

////////////////////////////////////////////////////////////////////////////////
#ifdef CUDA
void CORE_dtsmqr_acc(
    PLASMA_enum side, PLASMA_enum trans,
    int m1, int n1, int m2, int n2, int k, int ib,
    double *d_A1, int lda1,
    double *d_A2, int lda2,
    const double *d_V, int ldv,
    const double *d_T, int ldt,
    double *d_work, int ldwork,
    cublasHandle_t cublas_handle)
{
    int nb = m1;
    double *d_work2 = d_work + ib*nb;

    int j;
    for (j = 0; j < nb; j+=ib)
    {
        // W2 <= A1
        double done = 1.0;
        double dzero = 0.0;
        cublasDgeam(cublas_handle,
            CUBLAS_OP_N, CUBLAS_OP_N,
            ib, nb,
            &done,  &d_A1[j], lda1,
            &dzero,  d_A2,    lda2,
                     d_work2, ldwork);

        // W2 = W2 + V' A2
        cublasDgemm(cublas_handle,
            CUBLAS_OP_T, CUBLAS_OP_N,
            ib, nb, nb,
            &done, &d_V[j*ldv], ldv,
                   d_A2      , lda2,
            &done, d_work2   , ldwork);

        // W = T' W2
        cublasDgemm(cublas_handle,
            CUBLAS_OP_T, CUBLAS_OP_N,
            ib, nb, ib,
            &done, &d_T[j*ldt], ldt,
                    d_work2   , ldwork,
            &dzero, d_work    , ldwork);

        // A2 = A2 - V W
        double dmone = -1.0;
        cublasDgemm(cublas_handle,
            CUBLAS_OP_N, CUBLAS_OP_N,
            nb, nb, ib,
            &dmone, &d_V[j*ldv], ldv,
                     d_work    , ldwork,
            &done,   d_A2      , lda2);

        /* A1 = A1 - W */
        // W2 <- A1
        cublasDgeam(cublas_handle,
            CUBLAS_OP_N, CUBLAS_OP_N,
            ib, nb,
            &done,  &d_A1[j], lda1,
            &dzero,  d_A2,    lda2,
                     d_work2, ldwork);
        // A1 = W2 - W
        cublasDgeam(cublas_handle,
            CUBLAS_OP_N, CUBLAS_OP_N,
            ib, nb,
            &done,  d_work2,  ldwork,
            &dmone, d_work,   ldwork,
                    &d_A1[j], lda1);
    }
}
#endif

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Call PLASMA tile QR factorization.
 * Make a single-node multi-core PLASMA tile QR run.
 * Set the parameters and initialize the matrix to match the VSA run.
 *
 * @param nb matrix tile size
 * @param ib QR inner bocking
 * @param mt matrix height (number of tiles)
 * @param nt matrix width (number of tiles)
 * @param num_threads number of threads
 */
double plasma_tile_qr(int nb, int ib, int mt, int nt, int num_threads)
{
    PLASMA_Init(num_threads);
    PLASMA_desc *descA;
    PLASMA_desc *descT;
    PLASMA_Disable(PLASMA_AUTOTUNING);
    PLASMA_Set(PLASMA_TILE_SIZE, nb);
    PLASMA_Set(PLASMA_INNER_BLOCK_SIZE, ib);
    double *A = (double*)malloc(mt*nt*nb*nb*sizeof(double));
    int i;
    int j;
    for (j = 0; j < mt; j++)
        for (i = 0; i < nt; i++)
            pulsar_tile_init(&A[(i*mt+j)*nb*nb], j, i, nb);
    PLASMA_Desc_Create(
        &descA, A, PlasmaRealDouble,
        nb, nb, nb*nb, mt*nb, nt*nb, 0, 0, mt*nb, nt*nb);
    PLASMA_Alloc_Workspace_dgeqrf_Tile(mt*nb, nt*nb, &descT);
    double start = pulsar_time_of_day();
    PLASMA_dgeqrf_Tile(descA, descT);
    double stop = pulsar_time_of_day();
    pulsar_tile_print(&A[((nt-1)*mt+(mt-1))*nb*nb], nb);
    PLASMA_Desc_Destroy(&descA);
    free(A);
    PLASMA_Finalize();
    return (stop-start);
}
