/**
 * @file icl_deque.c
 * @brief Thread-safe double-ended queue.
 * @author Jakub Kurzak
 *
 * Implemented by protecting access to icl_list using spinlocks.
 * Also, unlike icl_list, icl_deque keeps track of its size.
 *
 * PULSAR Runtime
 * http://icl.utk.edu/pulsar/
 * Copyright (C) 2012-2015 University of Tennessee.
 */

#include "icl_deque.h"

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Creates a new deque.
 *
 * @returns A new deque. NULL on error.
 */
icl_deque_t* icl_deque_new()
{
    icl_deque_t *deque = (icl_deque_t*)malloc(sizeof(icl_deque_t));
    if (deque == NULL)
        return NULL;

    deque->list = icl_list_new();
    if (deque->list == NULL) {
        free(deque);
        return NULL;
    }
    if (pthread_spin_init(&deque->spinlock, PTHREAD_PROCESS_PRIVATE) != 0) {
        free(deque->list);
        free(deque);
        return NULL;
    }
    deque->size = 0;
    return deque;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Destroys a deque.
 *
 * @param deque -- The deque to destroy.
 * @param free_func -- The function that frees the node's data.
 *
 * @retval  0 on success.
 * @retval -1 on failure.
 */
int icl_deque_destroy(icl_deque_t *deque, void (*free_func)(void*))
{
    if (deque == NULL)
        return -1;

    if (pthread_spin_destroy(&deque->spinlock) != 0)
        return -1;

    if (icl_list_destroy(deque->list, free_func) != 0)
        return -1;

    free(deque);
    return 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Returns the first node in a deque.
 *
 * @param deque -- The deque to fetch from.
 *
 * @returns The node at the front of the deque. NULL if empty or error.
 */
icl_node_t* icl_deque_first(icl_deque_t *deque)
{
    if (deque == NULL)
        return NULL;

    pthread_spin_lock(&deque->spinlock);
    icl_node_t *node = icl_list_first(deque->list);
    pthread_spin_unlock(&deque->spinlock);

    return node;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Returns next node in a deque.
 *
 * @param deque -- The deque to fetch from.
 * @param node -- The node current node.
 *
 * @returns The next node. NULL if empty or error.
 */
icl_node_t* icl_deque_next(icl_deque_t *deque, icl_node_t *node)
{
    if (deque == NULL)
        return NULL;

    pthread_spin_lock(&deque->spinlock);
    icl_node_t *next = icl_list_next(deque->list, node);
    pthread_spin_unlock(&deque->spinlock);

    return next;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Inserts a node at the end of a deque.
 *
 * @param deque -- The deque to append to.
 * @param data -- The data to append.
 *
 * @returns The new node. NULL on error.
 */
icl_node_t* icl_deque_append(icl_deque_t *deque, void *data)
{
    if (deque == NULL || data == NULL)
        return NULL;

    pthread_spin_lock(&deque->spinlock);
    icl_node_t *node = icl_list_append(deque->list, data);
    deque->size++;
    pthread_spin_unlock(&deque->spinlock);

    return node;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Inserts a node at the front of a deque.
 *
 * @param deque -- The deque to prepend to.
 * @param data -- The data to prepent.
 *
 * @returns The new node. NULL on error.
 */
icl_node_t* icl_deque_prepend(icl_deque_t *deque, void *data)
{
    if (deque == NULL || data == NULL)
        return NULL;

    pthread_spin_lock(&deque->spinlock);
    icl_node_t *node = icl_list_prepend(deque->list, data);
    deque->size++;
    pthread_spin_unlock(&deque->spinlock);

    return node;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Deletes a node from a deque.
 *
 * @param deque -- The deque to delete from.
 * @param node -- The node to delete.
 * @param free_func -- The function that frees the node's data.
 *
 * @retval  0 on success.
 * @retval -1 on failure.
 */
int icl_deque_delete(
    icl_deque_t *deque, icl_node_t *node, void (*free_func)(void*))
{
    if (deque == NULL || node == NULL)
        return -1;

    pthread_spin_lock(&deque->spinlock);
    if (icl_list_delete(deque->list, node, free_func) == 0) {
        deque->size--;
        pthread_spin_unlock(&deque->spinlock);
        return 0;
    }
    else {
        pthread_spin_unlock(&deque->spinlock);
        return -1;
    }
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Returns the size of a deque.
 *
 * @param deque -- The deque to get size of.
 *
 * @returns -- The size of the deque. -1 on error.
 */
int icl_deque_size(icl_deque_t *deque)
{
    if (deque == NULL)
        return -1;

    pthread_spin_lock(&deque->spinlock);
    int size = deque->size;
    pthread_spin_unlock(&deque->spinlock);
    return size;
}
