/**
 * @file icl_hash.c
 * @brief Dependency-free hash table.
 * @author Keith Seymour
 *
 * PULSAR Runtime
 * http://icl.utk.edu/pulsar/
 * Copyright (C) 2012-2015 University of Tennessee.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <limits.h>

#include "icl_hash.h"

#define BITS_IN_int     (sizeof(int) * CHAR_BIT)
#define THREE_QUARTERS  ((int)((BITS_IN_int * 3) / 4))
#define ONE_EIGHTH      ((int)(BITS_IN_int / 8))
#define HIGH_BITS       (~((unsigned int)(~0) >> ONE_EIGHTH))

static unsigned int hash_pjw(void *key);
static int string_compare(void *a, void *b);

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief A simple string hash.
 *
 * An adaptation of Peter Weinberger's (PJW) generic hashing algorithm
 * based on Allen Holub's version.
 * Accepts a pointer to a datum to be hashed and returns an unsigned integer.
 *
 * @param key -- The string to be hashed.
 *
 * @returns The hash index.
 */
static unsigned int hash_pjw(void *key)
{
    char *datum = (char*)key;
    unsigned int hash_value, i;

    if(!datum)
        return 0;

    for (hash_value = 0; *datum; ++datum) {
        hash_value = (hash_value << ONE_EIGHTH) + *datum;
        if ((i = hash_value & HIGH_BITS) != 0)
            hash_value = (hash_value ^ (i >> THREE_QUARTERS)) & ~HIGH_BITS;
    }
    return (hash_value);
}

////////////////////////////////////////////////////////////////////////////////
static int string_compare(void *a, void *b)
{
    return (strcmp( (char*)a, (char*)b ) == 0);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Creates a new hash table.
 *
 * @param nbuckets -- The number of buckets to create.
 * @param hash_function -- The pointer to the hashing function.
 * @param hash_key_compare -- The pointer to the hash key comparison function.
 *
 * @returns A pointer to new hash table.
 */
icl_hash_t* icl_hash_create(
    int nbuckets,
    unsigned int (*hash_function)(void*),
    int (*hash_key_compare)(void*, void*))
{
    icl_hash_t *ht;
    int i;

    ht = (icl_hash_t*)malloc(sizeof(icl_hash_t));
    if(!ht)
        return NULL;

    ht->nentries = 0;
    ht->buckets = (icl_entry_t**)malloc(nbuckets*sizeof(icl_entry_t*));
    if(!ht->buckets) {
        free(ht);
        return NULL;
    }
    ht->nbuckets = nbuckets;
    for(i = 0; i < ht->nbuckets; i++)
        ht->buckets[i] = NULL;

    ht->hash_function = hash_function ? hash_function : hash_pjw;
    ht->hash_key_compare = hash_key_compare ? hash_key_compare : string_compare;

    return ht;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Searches for an entry in a hash table.
 *
 * @param ht -- The hash table to be searched.
 * @param key -- The key of the item to search for.
 *
 * @returns A pointer to the data corresponding to the key.
 * NULL if the key is not found.
 */
void *icl_hash_find(icl_hash_t *ht, void* key)
{
    icl_entry_t* curr;
    unsigned int hash_val;

    if(!ht || !key)
        return NULL;

    hash_val = (*ht->hash_function)(key) % ht->nbuckets;
    for (curr = ht->buckets[hash_val]; curr != NULL; curr = curr->next)
        if (ht->hash_key_compare(curr->key, key))
            return(curr->data);

    return NULL;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Inserts an item into a hash table.
 *
 * @param ht -- The hash table.
 * @param key -- The key of the new item.
 * @param data -- The pointer to the new item's data.
 *
 * @returns A pointer to the new item. NULL on error.
 */
icl_entry_t* icl_hash_insert(icl_hash_t *ht, void* key, void *data)
{
    icl_entry_t *curr;
    unsigned int hash_val;

    if(!ht || !key)
        return NULL;

    hash_val = (*ht->hash_function)(key) % ht->nbuckets;
    for (curr = ht->buckets[hash_val]; curr != NULL; curr = curr->next)
        if (ht->hash_key_compare(curr->key, key))
            return(NULL); // The key already exists.

    // IF key was not found.
    curr = (icl_entry_t*)malloc(sizeof(icl_entry_t));
    if(!curr)
        return NULL;

    curr->key = key;
    curr->data = data;
    curr->next = ht->buckets[hash_val]; // Add at the start.

    ht->buckets[hash_val] = curr;
    ht->nentries++;

    return curr;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Replaces an entry in a hash table with a given entry.
 *
 * @param ht -- The hash table.
 * @param key -- The key of the new item.
 * @param data -- The pointer to the new item's data.
 * @param olddata -- The pointer to the old item's data (set upon return).
 *
 * @returns A pointer to the new item. NULL on error.
 */
icl_entry_t *icl_hash_update_insert(
    icl_hash_t *ht, void* key, void *data, void **olddata)
{
    icl_entry_t *curr, *prev;
    unsigned int hash_val;

    if(!ht || !key)
        return NULL;

    hash_val = (*ht->hash_function)(key) % ht->nbuckets;

    // Scan bucket[hash_val] for the key.
    for (prev = NULL, curr=ht->buckets[hash_val];
         curr != NULL;
         prev = curr, curr = curr->next)
        // IF key found, remove the node from the list,
        // free the old key, and setup olddata for the return.
        if (ht->hash_key_compare(curr->key, key)) {
            if (olddata != NULL) {
                *olddata = curr->data;
                free(curr->key);
            }
            if (prev == NULL)
                ht->buckets[hash_val] = curr->next;
            else
                prev->next = curr->next;
        }

    // Since the key was either not found,
    // or found-and-removed, create and prepend a new node.
    curr = (icl_entry_t*)malloc(sizeof(icl_entry_t));
    if(curr == NULL)
        return NULL; // Out of memory.

    curr->key = key;
    curr->data = data;
    curr->next = ht->buckets[hash_val]; // Add at the start.

    ht->buckets[hash_val] = curr;
    ht->nentries++;

    if (olddata != NULL && *olddata != NULL)
        *olddata = NULL;

    return curr;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Frees one hash table entry located by a key.
 * Key and data are freed using functions.
 *
 * @param ht -- The hash table.
 * @param key -- The key of the item to be deleted.
 * @param free_key -- The pointer to the function that frees the key.
 * @param free_data -- The pointer to the function that frees the data.
 *
 * @retval 0 on success.
 * @retval -1 on failure.
 */
int icl_hash_delete(
    icl_hash_t *ht, void* key,
    void (*free_key)(void*),
    void (*free_data)(void*))
{
    icl_entry_t *curr, *prev;
    unsigned int hash_val;

    if(!ht || !key)
        return -1;

    hash_val = (*ht->hash_function)(key) % ht->nbuckets;

    prev = NULL;
    for (curr = ht->buckets[hash_val]; curr != NULL;) {
        if (ht->hash_key_compare(curr->key, key)) {
            if (prev == NULL) {
                ht->buckets[hash_val] = curr->next;
            }
            else {
                prev->next = curr->next;
            }
            if (*free_key && curr->key)
                (*free_key)(curr->key);
            if (*free_data && curr->data)
                (*free_data)(curr->data);
            ht->nentries++;
            free(curr);
            return 0;
        }
        prev = curr;
        curr = curr->next;
    }
    return -1;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Destroys a hash table.
 * Keys and data are freed using functions.
 *
 * @param ht -- The hash table to destroy.
 * @param free_key -- The pointer to function that frees the keys.
 * @param free_data -- The pointer to function that frees the data.
 *
 * @retval 0 on success.
 * @retval -1 on failure.
 */
int icl_hash_destroy(
    icl_hash_t *ht,
    void (*free_key)(void*),
    void (*free_data)(void*))
{
    icl_entry_t *bucket, *curr, *next;
    int i;

    if(!ht)
        return -1;

    for (i = 0; i < ht->nbuckets; i++) {
        bucket = ht->buckets[i];
        for (curr = bucket; curr!=NULL;) {
            next = curr->next;
            if (*free_key && curr->key)
                (*free_key)(curr->key);
            if (*free_data && curr->data)
                (*free_data)(curr->data);
            free(curr);
            curr = next;
        }
    }
    if (ht->buckets)
        free(ht->buckets);
    if (ht)
        free(ht);

    return 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Dumps the hash table's contents to the given file pointer.
 *
 * @param stream -- The file to dump the hash table to.
 * @param ht -- The hash table to be dumped.
 *
 * @retval 0 on success.
 * @retval -1 on failure.
 */
int icl_hash_dump(FILE *stream, icl_hash_t *ht)
{
    icl_entry_t *bucket, *curr;
    int i;

    if(!ht)
        return -1;

    for (i = 0; i < ht->nbuckets; i++) {
        bucket = ht->buckets[i];
        for (curr = bucket; curr!=NULL;) {
            if (curr->key)
                fprintf(stream,
                    "icl_hash_dump: %s: %p\n", (char*)curr->key, curr->data);
            curr=curr->next;
        }
    }
    return 0;
}
