/**
 * @file icl_list.c
 * @brief Dependency-free linked list.
 * @author Keith Seymour
 *
 * PULSAR Runtime
 * http://icl.utk.edu/pulsar/
 * Copyright (C) 2012-2015 University of Tennessee.
 */

#include <stdio.h>
#include <stdlib.h>

#include "icl_list.h"

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Creates a new linked list.
 *
 * @returns A new linked list. NULL on error.
 */
icl_list_t* icl_list_new()
{
    icl_list_t *node;

    node = (icl_list_t*)malloc(sizeof(icl_list_t));
    if (!node)
        return NULL;

    node->flink = NULL;
    node->blink = node;
    node->data = NULL;

    return(node);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Inserts a new node after the specified node.
 *
 * @param head -- The linked list.
 * @param pos -- The insertion position (the node to append to).
 * @param data -- The pointer to the data to be inserted.
 *
 * @returns A pointer to the new node. NULL on error.
 */
icl_list_t* icl_list_insert(icl_list_t *head, icl_list_t *pos, void *data)
{
    icl_list_t *node;

    if(!head || !pos)
        return NULL;

    node = (icl_list_t*)malloc(sizeof(icl_list_t));
    if (!node)
        return NULL;

    node->blink = pos;
    node->flink = pos->flink;
    node->data = data;

    if (pos->flink)
        pos->flink->blink = node;
    else
        head->blink = node; // The node at end of the list.

    pos->flink = node;
    return node;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Deletes the specified node.
 *
 * @param head -- The linked list containing the node to be deleted.
 * @param pos -- The node to be deleted.
 * @param free_function -- The function that frees the node's data.
 *
 * @retval 0 on success.
 * @retval -1 on failure.
 */
int icl_list_delete(
    icl_list_t *head, icl_list_t *pos, void (*free_function)(void *))
{
    if (!pos || !head)
        return -1;
    if (pos == head)
        return -1;

    if (free_function && pos->data)
        (*free_function)(pos->data);

    pos->blink->flink = pos->flink;

    if (pos->flink)
        pos->flink->blink = pos->blink;
    else
        head->blink = pos->blink; // Position at the end of the list.

    free(pos);
    return 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Finds a data item in a linked list.
 *
 * @param head -- The linked list.
 * @param data -- The data to be found.
 * @param compare -- The function that compares the data items.
 *
 * @returns A pointer to the node, if found. Otherwise NULL.
 */
icl_list_t* icl_list_search(
    icl_list_t *head, void *data, int (*compare)(void*, void*))
{
    icl_list_t *node;

    if (!head)
        return NULL;

    for (node = head->flink; node != NULL; node = node->flink) {
        if (!node->data)
            continue;
        if ((compare && (*compare)(node->data, data) == 0))
            break;
        else if (node->data == data)
            break; // compare == NULL, then direct comparison of pointers.
    }
    return(node);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Inserts data into a sorted list.
 * Does not support direct comparison of pointers.
 *
 * @param head -- The linked list.
 * @param data -- The data to be inserted.
 * @param compare -- The function that compares the data items.
 *
 * @returns A pointer to the new node. NULL on error.
 */
icl_list_t* icl_list_isort(
    icl_list_t *head, void *data, int (*compare)(void*, void*))
{
    if (!head || !compare)
        return NULL;

    if (head->flink == NULL)
        return (icl_list_insert(head, head, data));

    icl_list_t *node = head->flink;
    if (compare(data, node->data) <= 0)
        return (icl_list_insert(head, head, data));

    while (node->flink != NULL && compare(data, node->flink->data) > 0)
        node = node->flink;
    return (icl_list_insert(head, node, data));
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Destroys a linked list.
 *
 * @param head -- The linked list to be destroyed.
 * @param free_function -- The function that frees the node's data.
 *
 * @retval 0 on success.
 * @retval -1 on failure.
 */
int
icl_list_destroy(icl_list_t *head, void (*free_function)(void*))
{
    icl_list_t *node, *tmpnode;

    if (!head)
        return -1;

    for (node = head; node != NULL;) {
        tmpnode = node->flink;

        if (free_function != NULL && node->data != NULL)
            (*free_function)(node->data);

        free(node);
        node = tmpnode;
    }
    return 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Returns the number of items in a linked list.
 *
 * @param head -- The linked list.
 *
 * @returns The number of items in the list. -1 on error.
 */
int icl_list_size(icl_list_t *head)
{
    int size = 0;

    if (!head)
        return -1;

    while ((head=head->flink))
        size++;

    return size;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Returns the first item in a linked list.
 *
 * @param head -- The linked list.
 *
 * @returns A pointer to the first item. NULL on error.
 */
icl_list_t* icl_list_first(icl_list_t *head)
{
    if (head)
        return head->flink;

    return NULL;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Returns the last item in a linked list.
 *
 * @param head -- The linked list.
 *
 * @returns A pointer to the last item. NULL on error.
 */
icl_list_t* icl_list_last(icl_list_t *head)
{
    if (head)
        return head->blink;

    return NULL;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Returns the node following the specified node.
 *
 * @param head -- The list containing the specified node.
 * @param pos -- The node whose successor should be returned.
 *
 * @returns A pointer to the next node. NULL on error.
 */
icl_list_t* icl_list_next(icl_list_t *head, icl_list_t *pos)
{
    if(pos)
        return pos->flink;

    return NULL;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Returns the node preceding the specified node.
 *
 * @param head -- The list containing the specified node.
 * @param pos -- The node whose predecessor should be returned.
 *
 * @returns A pointer to the previous node. NULL on error.
 */
icl_list_t* icl_list_prev(icl_list_t *head, icl_list_t *pos)
{
    if (pos && pos->blink != NULL && pos != head &&
        pos->blink != head && pos->blink != pos)
        return pos->blink;

    return NULL;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Concatenates two linked lists.
 *
 * @param head1 -- The first linked list.
 * @param head2 -- The second linked list.
 *
 * @returns A pointer to the new linked list, which consists of <head1,head2>.
 * NULL on error.
 */
icl_list_t *icl_list_concat(icl_list_t *head1, icl_list_t *head2)
{
    if (!head1 || !head2 || !head1->blink || !head2->flink)
        return NULL;

    head1->blink->flink = head2->flink;
    head2->flink->blink = head1->blink;
    head1->blink = head2->blink;

    free(head2);
    return(head1);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Inserts a node at the beginning of a list.
 *
 * @param head -- The linked list.
 * @param data -- The data to be inserted.
 *
 * @returns A pointer to the new node. NULL on error.
 */
icl_list_t *icl_list_prepend(icl_list_t *head, void *data)
{
    return (icl_list_insert(head, head, data));
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Inserts a node at the end of a list.
 *
 * @param head -- The linked list.
 * @param data -- The data to be inserted.
 *
 * @returns A pointer to the new node. NULL on error.
 */
icl_list_t* icl_list_append(icl_list_t *head, void *data)
{
    return (icl_list_insert(head, head->blink, data));
}
