/**
 * @file prt_channel.h
 * @brief PRT data channel.
 * @author Jakub Kurzak
 *
 * PULSAR Runtime
 * http://icl.utk.edu/pulsar/
 * Copyright (C) 2012-2015 University of Tennessee.
 */

#ifndef PRT_CHANNEL_H
#define PRT_CHANNEL_H

#include "prt.h"

#ifdef __cplusplus
extern "C" {
#endif

struct prt_vdp_s;
struct prt_packet_s;

/**
 * @brief VDP's data channel.
 * Implements a data link between a pair of VDPs.
 * Identifies the source and destination VDPs by tuples.
 * Contains a thread-safe list of data packets.
 *
 * The in_stream is used when the recipient device pulls:
 * host->device, device->device (second stage).
 * The out_stream is used when the sender device pushes:
 * device->host, device->device (first stage).
 */
typedef struct prt_channel_s {
    struct prt_vdp_s *dst_vdp;  // Back pointer to the destination VDP.
    struct prt_vdp_s *src_vdp;  // Back pointer to the source VDP.
    struct prt_proxy_s *proxy;  // Pointer to the comm proxy.
    size_t size;                // Max packet size in bytes.
    int *src_tuple;             // Tuple of the source VDP.
    int src_slot;               // Slot in the source VDP.
    int *dst_tuple;             // Tuple of the destination VDP.
    int dst_slot;               // Slot in the destination VDP.
    int src_node;               // Source node.
    int dst_node;               // Destination node.
    int tag;                    // Tag for MPI communication.
    icl_deque_t *packets;       // List of data packets.
    int active;                 // Activity flag.
    cudaStream_t in_stream;     // Inbound stream.
    cudaStream_t out_stream;    // Outbound stream.
} prt_channel_t;

/**
 * @brief VDP's data channel direction.
 * Identifies the direction of a VDP channel during insertion.
 */
typedef enum prt_channel_direction_e {
    PRT_INPUT_CHANNEL, PRT_OUTPUT_CHANNEL
} prt_channel_direction_t;

prt_channel_t* prt_channel_new(
    size_t size,
    int *src_tuple, int src_slot,
    int *dst_tuple, int dst_slot);
void prt_channel_delete(prt_channel_t *channel);
void prt_channel_push_host(
    struct prt_vdp_s *vdp, prt_channel_t *channel, struct prt_packet_s *packet);
void prt_channel_push_device(
    struct prt_vdp_s *vdp, prt_channel_t *channel, struct prt_packet_s *packet);
struct prt_packet_s *prt_channel_pop(prt_channel_t *channel);
int prt_channel_empty(prt_channel_t *channel);
int prt_channel_compare(void *channel1, void *channel2);
void prt_channel_off(prt_channel_t *channel);
void prt_channel_on(prt_channel_t *channel);

#ifdef __cplusplus
}
#endif

#endif /* PRT_CHANNEL_H */
