/**
 * @file prt_device.c
 * @brief PRT device.
 * @author Jakub Kurzak
 *
 * PULSAR Runtime
 * http://icl.utk.edu/pulsar/
 * Copyright (C) 2012-2015 University of Tennessee.
 */

#include "prt_device.h"

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Creates a new device.
 *
 * @param rank -- The local rank of the device.
 * @param accelerator -- The global rank of the device.
 * @param agent_rank -- The rank of the communication agent.
 *
 * @returns A new device object.
 */
prt_device_t* prt_device_new(int rank, int accelerator, int agent_rank)
{
    // Allocate the device.
    prt_device_t *device = (prt_device_t*)malloc(sizeof(prt_device_t));
    prt_assert(device != NULL, "malloc failed");

    // Initialize the device.
    device->rank = rank;
    device->accelerator = accelerator;
    device->finished = 0;
    device->node = NULL;

    // Create the list of VDPs.
    device->vdps = icl_list_new();
    prt_assert(device->vdps != NULL, "icl_list_new failed");

    // Return the device.
    return device;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Destroys a device.
 *
 * @param device -- The device to destroy.
 */
void prt_device_delete(prt_device_t *device)
{
    // Destroy the list of VDPs.
    int size = icl_list_size(device->vdps);
    prt_assert(size == 0, "destroying non-empty list");
    int status = icl_list_destroy(device->vdps, NULL);
    prt_assert(status == 0, "icl_list_destroy failed");

    // Free the device.
    free(device);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Implements device processing cycle.
 *
 * @param device -- The device to cycle.
 */
void prt_device_cycle(prt_device_t *device)
{
    // Return if finished.
    if (device->finished)
        return;

    // Assign node if NULL.
    if (device->node == NULL) {
        device->node = icl_list_first(device->vdps);
        // Return if list empty.
        if (device->node == NULL) {
            device->finished = 1;
            return;
        }
    }
    // Set the device.
    cudaError_t error = cudaSetDevice(device->rank);
    prt_assert(error == cudaSuccess, cudaGetErrorString(error));

    prt_vdp_t *vdp = (prt_vdp_t*)device->node->data;
    // IF not finished and inputs ready.
    if (vdp->counter > 0 && prt_vdp_ready(vdp)) {
        // Fire and decrement.
        svg_trace_start_gpu(vdp->stream);
        vdp->function(vdp);
        svg_trace_stop_gpu(vdp->stream, vdp->color);
        vdp->counter--;
    }
    // IF finished or lazy scheduling.
    if (vdp->counter == 0 ||
        device->vsa->config->vdp_scheduling == PRT_VDP_SCHEDULING_LAZY) {
        // Advance.
        icl_list_t *prev = device->node;
        device->node = icl_list_next(device->vdps, device->node);
        // IF burned out.
        if (vdp->counter == 0)
            // Remove the VDP from the list
            // Do not destroy the VDP itself.
            icl_list_delete(device->vdps, prev, NULL);
    }
}
