/**
 * @file prt_packet.h
 * @brief PRT data packet.
 * @author Jakub Kurzak
 *
 * PULSAR Runtime
 * http://icl.utk.edu/pulsar/
 * Copyright (C) 2012-2015 University of Tennessee.
 */

#ifndef PRT_PACKET_H
#define PRT_PACKET_H

#include "prt.h"

#ifdef __cplusplus
extern "C" {
#endif

struct prt_vdp_s;
struct prt_channel_s;
struct gpu_malloc_s;

/**
 * @brief VDP's data packet
 * A packet of data transferred through VDP's channels.
 *
 * "num_refs" is a multi-access synchronization variable.
 * Therefore, declared as volatile and accessed with atomics.
 */
typedef struct prt_packet_s {
    void *data;                     // Pointer to the data.
    size_t size;                    // Data size in bytes.
    volatile int num_refs;          // Number of active references.
    prt_location_t location;        // Packet location.
    int device_rank;                // Device number.
    struct gpu_malloc_s *devmem;    // Device memory allocator.
} prt_packet_t;

prt_packet_t *prt_packet_new_host(size_t size, void *data);
prt_packet_t *prt_packet_new_device(
    size_t size, void *data, struct prt_vdp_s *vdp);
void prt_packet_resize_host(prt_packet_t *packet, size_t size);
void prt_packet_release_host(prt_packet_t *packet);
void prt_packet_release_device(prt_packet_t *packet);

void prt_packet_host_to_device(
    prt_packet_t *src_packet, struct prt_channel_s *channel);
void prt_packet_device_to_host(
    prt_packet_t *src_packet, struct prt_channel_s *channel);
void prt_packet_device_to_device(
    prt_packet_t *src_packet, struct prt_channel_s *channel);
void prt_packet_device_to_device_direct(
    prt_packet_t *src_packet, struct prt_channel_s *channel);
void prt_packet_device_mpi_to_host(
    prt_packet_t *src_packet, struct prt_channel_s *channel, int agent);

#ifdef __cplusplus
}
#endif

#endif /* PRT_PACKET_H */
