/**
 * @file prt_proxy.h
 * @brief PRT communication proxy.
 * @author Jakub Kurzak
 *
 * PULSAR Runtime
 * http://icl.utk.edu/pulsar/
 * Copyright (C) 2012-2015 University of Tennessee.
 */

#ifndef PRT_PROXY_H
#define PRT_PROXY_H

#include "prt.h"

#ifdef __cplusplus
extern "C" {
#endif

struct prt_vsa_s;
struct prt_packet_s;
struct prt_channel_s;
struct prt_request_s;

/**
 * @brief Maximum tags per node.
 * Size of the proxy's hash table for tags.
 * It should be a prime number.
 */
#define PRT_PROXY_MAX_TAGS_PER_NODE 10003

/**
 * @brief Maximum numbers of outstanding MPI send requests per agent.
 */
#define PRT_PROXY_MAX_SENDS_PER_AGENT 1

/**
 * @brief Maximum numbers of outstanding MPI receive requests per agent.
 */
 #define PRT_PROXY_MAX_RECVS_PER_AGENT 1

/**
 * @brief VSA's proxy.
 *
 * The reason for the num_callbacks counter is the following:
 * Empty transfers queue does not mean there is nothing pending.
 * Communication requests may be sitting in a stream waiting to be queued.
 */
typedef struct prt_proxy_s {
    struct prt_vsa_s *vsa;          // Back pointer to the parent VSA.
    int num_agents;                 // Number of agents (threads + devices).
    icl_hash_t *tags_hash;          // Channel lookup by rank-tag.
    icl_deque_t **sends_requested;  // Sends requested by channels (per agent).
    icl_list_t **sends_posted;      // Sends posted, not completed (per agent).
    icl_list_t *recvs_posted;       // Recvs posted, not completed (all agents).
    icl_deque_t *transfers;         // Requested accelerator transfers.
    size_t max_channel_size;        // Maximum size of a channel in bytes.
    volatile int num_callbacks;     // Number of pending callbacks.
} prt_proxy_t;

prt_proxy_t* prt_proxy_new(int num_agents);
void prt_proxy_delete(prt_proxy_t *proxy);
void prt_proxy_max_channel_size(prt_proxy_t *proxy, struct prt_channel_s *channel);
void prt_proxy_recv(prt_proxy_t *proxy, struct prt_request_s *request);
void prt_proxy_mpi(prt_proxy_t *proxy);
void prt_proxy_cuda(prt_proxy_t *proxy);
double prt_proxy_run(prt_proxy_t *proxy);

#ifdef __cplusplus
}
#endif

#endif /* PRT_PROXY_H */
