/**
 * @file prt_request.c
 * @brief PRT communication request.
 * @author Jakub Kurzak
 *
 * PULSAR Runtime
 * http://icl.utk.edu/pulsar/
 * Copyright (C) 2012-2015 University of Tennessee.
 */

#include "prt_request.h"

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Creates a new request.
 *
 * @param packet -- The packet to create the request for.
 * @param count -- The number of data items.
 * @param datatype -- The type of data items.
 * @param peer -- The peer communicating node.
 * @param tag -- The MPI tag of the message.
 *
 * @returns A new request.
 */
prt_request_t *prt_request_new(
    prt_packet_t *packet, size_t size, int peer, int tag)
{
    prt_request_t *request = (prt_request_t*)malloc(sizeof(prt_request_t));
    prt_assert(request != NULL, "malloc failed");
    request->packet = packet;
    request->size = size;
    request->peer = peer;
    request->tag = tag;
    return request;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Destroys a request.
 *
 * @param request -- The request to destroy.
 */
void prt_request_delete(prt_request_t *request)
{
    free(request);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Posts a send request.
 *
 * @param request -- The send request to post.
 */
void prt_request_send(prt_request_t *request)
{
    int retval;
    // Post the request.
    svg_trace_start_cpu(0);
    retval = MPI_Isend(
        request->packet->data,
        request->size,
        MPI_BYTE,
        request->peer,
        request->tag,
        MPI_COMM_WORLD,
        &request->request);
    svg_trace_stop_cpu(0, -DarkKhaki);
    prt_assert(retval == MPI_SUCCESS, "error in MPI_Isend");
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Posts a receive request.
 *
 * @param request -- The receive request to post.
 */
void prt_request_recv(prt_request_t *request)
{
    int retval;
    // Post the request.
    svg_trace_start_cpu(0);
    retval = MPI_Irecv(
        request->packet->data,
        request->size,
        MPI_BYTE,
        request->peer,
        request->tag,
        MPI_COMM_WORLD,
        &request->request);
    svg_trace_stop_cpu(0, -OliveDrab);
    prt_assert(retval == MPI_SUCCESS, "error in MPI_Isend");
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Tests a request.
 * Traces only completed requests.
 *
 * @param request -- The request to test.
 *
 * @retval 1 if operation completed.
 * @retval 0 if operation is pending.
 */
int prt_request_test(prt_request_t *request)
{
    int flag;
    int retval;
    // Test the request.
    // Check the status.
    // Trace if completed.
    svg_trace_start_cpu(0);
    retval = MPI_Test(&request->request, &flag, &request->status);
    if (flag == 1)
        svg_trace_stop_cpu(0, -RosyBrown);
    prt_assert(retval == MPI_SUCCESS, "error in MPI_Test");
    return flag;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Cancels a request.
 * Cancels them MPI request, releases the packet, frees the request object.
 *
 * @param request -- The request to cancel.
 */
void prt_request_cancel(prt_request_t *request)
{
    int retval;
    // Cancel the request.
    retval = MPI_Cancel(&request->request);
    prt_assert(retval == MPI_SUCCESS, "error in MPI_Cancel");

    // Release the packet.
    prt_packet_release_host(request->packet);

    // Free the request.
    free(request);
}
