/**
 * @file prt_tuple.c
 * @brief PRT tuple.
 * @author Jakub Kurzak
 *
 * PULSAR Runtime
 * http://icl.utk.edu/pulsar/
 * Copyright (C) 2012-2015 University of Tennessee.
 */

#include "prt_tuple.h"

static unsigned int fnv_hash(void *key, int len);

////////////////////////////////////////////////////////////////////////////////
/**
 * @ingroup API_core
 * @brief Creates a new tuple.
 * Allocates memory for the tuple plus the termination symbol (INT_MAX).
 * Fills out the tuple with the integers on the list.
 * There is also a set of macros, prt_tuple_new1/2/3/4/5/6,
 * where the length of the tuple is indicated by the number in the name.
 * Because this is such a tiny function, and is mostly intended to be accessed
 * through macros, skipping error checks for input parameters.
 *
 * @param len -- The length of the tuple.
 * @param ... -- A list of elements of type int.
 *
 * @returns A pointer to an array of integers terminated by INT_MAX.
 */
int *prt_tuple_new(int len, ...)
{
    int *tuple = (int*)malloc((len+1)*sizeof(int));
    prt_assert(tuple != NULL, "malloc failed");
    int i;
    va_list ap;
    va_start(ap, len);
    for (i = 0; i < len; i++)
        tuple[i] = va_arg(ap, int);
    va_end(ap);
    tuple[len] = INT_MAX;
    return tuple;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Returns the length of a tuple.
 *
 * @param tuple -- The tuple to return the length of.
 *
 * @returns The length of the tuple without the terminating symbol.
 */
int prt_tuple_len(int *tuple)
{
    int *p = tuple;
    while (*p++ != INT_MAX);
    return (p-tuple-1);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Concatenates a list of tuples.
 * Concatenates a variable-length, NULL-terminated, list of tuples.
 *
 * @param first_tuple -- The first tuple in the sequence.
 * @param ... -- A list of more tuples.
 *
 * @returns The aggregate tuple.
 */
int* prt_tuple_cat(int *first_tuple, ...)
{
    int len;
    int *tuple;
    va_list ap;
    va_start(ap, first_tuple);
    len = prt_tuple_len(first_tuple);
    while ((tuple = va_arg(ap, int*)) != NULL)
        len += prt_tuple_len(tuple);
    va_end(ap);

    int *out_tuple = (int*)malloc((len+1)*sizeof(int));
    prt_assert(out_tuple != NULL, "malloc failed");
    va_start(ap, first_tuple);
    while(*first_tuple != INT_MAX)
        *out_tuple++ = *first_tuple++;
    while ((tuple = va_arg(ap, int*)) != NULL)
        while(*tuple != INT_MAX)
            *out_tuple++ = *tuple++;
    va_end(ap);

    *out_tuple = INT_MAX;
    return (out_tuple-len);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Destroys a tuple.
 *
 * @param tuple -- The tuple to destroy.
 */
void prt_tuple_delete(int *tuple)
{
    // Free the tuple.
    free(tuple);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Copies a tuple.
 *
 * @param in_tuple -- The tuple to copy.
 *
 * @returns A new copy of the tuple.
 */
int *prt_tuple_copy(int *in_tuple)
{
    int i;
    int len = 0;
    int *out_tuple;
    while (in_tuple[len] != INT_MAX)
        len++;
    out_tuple = (int*)malloc((len+1)*sizeof(int));
    prt_assert(out_tuple != NULL, "malloc failed");
    for (i = 0; i < len; i++)
        out_tuple[i] = in_tuple[i];
    out_tuple[len] = INT_MAX;
    return out_tuple;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Compares two tuples.
 *
 * @param tuple_a -- The first tuple.
 * @param tuple_b -- The second tuple.
 *
 * @retval -1 if tuple_a is less than tuple_b.
 * @retval  0 if tuple_a is equal to tuple_b.
 * @retval  1 if tuple_a is greater than tuple_b.
 */
int prt_tuple_compare(void *tuple_a, void *tuple_b)
{
    int *tupa = (int*)tuple_a;
    int *tupb = (int*)tuple_b;
    while (*tupa != INT_MAX && *tupb != INT_MAX) {
        if (*tupa < *tupb) return -1;
        else if (*tupa > *tupb) return 1;
        tupa++;
        tupb++;
    }
    if (*tupa == INT_MAX && *tupb == INT_MAX) return 0;
    else if (*tupa == INT_MAX) return -1;
    else if (*tupb == INT_MAX) return 1;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Checks if two tuples are identical.
 * Check if tuples are identical in length and content.
 *
 * @param tuple_a -- The first tuple.
 * @param tuple_b -- The second tuple.
 *
 * @retval 0 if tuples are different.
 * @retval 1 if tuples are identical.
 */
int prt_tuple_equal(void *tuple_a, void *tuple_b)
{
    return (prt_tuple_compare(tuple_a, tuple_b) == 0);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Prints a tuple.
 *
 * @param tuple -- The tuple to print.
 */
void prt_tuple_print(int *tuple)
{
    while (*tuple != INT_MAX)
        printf("%3d", *tuple++);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Hashes a tuple.
 * This function is required by the VSA's tuples hash table.
 * It computes the lenght in characters and calls a string hashing function.
 *
 * @param tuple -- The tuple to hash.
 *
 * @returns hash
 */
unsigned int prt_tuple_hash(void *tuple)
{
    int len = 0;
    int *tup = (int*)tuple;
    while (tup[len] != INT_MAX)
        len++;
    return fnv_hash(tuple, len*sizeof(int));
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Implements the Fowler/Noll/Vo hashing function.
 * http://eternallyconfuzzled.com/tuts/algorithms/jsw_tut_hashing.aspx
 *
 * @param key -- The pointer to the string.
 * @param len -- The length of the string.
 *
 * @returns The hash of the key.
 */
static unsigned int fnv_hash(void *key, int len)
{
    unsigned char *p = key;
    unsigned h = 2166136261;
    int i;
    for (i = 0; i < len; i++)
        h = (h * 16777619) ^ p[i];
    return h;
}
