/**
 * @file prt_vdp.h
 * @brief Virtual Data Processor.
 * @author Jakub Kurzak
 *
 * PULSAR Runtime
 * http://icl.utk.edu/pulsar/
 * Copyright (C) 2012-2015 University of Tennessee.
 */

#ifndef PRT_VDP_H
#define PRT_VDP_H

#include "prt.h"

#ifdef __cplusplus
extern "C" {
#endif

struct prt_vdp_s;
struct prt_thread_s;
struct prt_channel_s;
enum prt_channel_direction_e;

/**
 * @brief VDP's function pointer.
 * Defines the type of the pointer to the VDP's function.
 */
typedef void (*prt_vdp_function_t)(struct prt_vdp_s*);

/**
 * @brief Virtual Data Processor (VDP).
 * Is uniquely identified by a tuple.
 * Fires for a predefined number of cycles.
 * Has a fixed number of input and output channels.
 * Has a persistent local store.
 * Has access to read-only global store.
 */
typedef struct prt_vdp_s {
    prt_location_t location;        // VDP's location (host or device).
    struct prt_thread_s *thread;    // Back pointer to the parent thread.
    struct prt_device_s *device;    // Back pointer to the parent device.
    struct prt_vsa_s *vsa;          // Back pointer to the VSA.
    int *tuple;                     // VDP's tuple.
    int counter;                    // VDP's countdown counter.
    int num_inputs;                 // Number of input channels.
    struct prt_channel_s **input;   // Array of input channels.
    int num_outputs;                // Number of output channels.
    struct prt_channel_s **output;  // Array of output channels.
    prt_vdp_function_t function;    // VDP's function.
    void *local_store;              // VDP's persistent local storage.
    void *global_store;             // VSA's read-only global store.
    int color;                      // RGB color for tracing.
    cudaStream_t stream;            // VDP's stream.
} prt_vdp_t;

prt_vdp_t* prt_vdp_new(
    int *tuple, int counter,
    prt_vdp_function_t function,
    size_t local_store_size,
    int num_inputs, int num_outputs, int color);
void prt_vdp_delete(prt_vdp_t *vdp);
void prt_vdp_annihilate(prt_vdp_t *vdp);
void prt_vdp_channel_insert(
    prt_vdp_t *vdp, struct prt_channel_s *channel,
    enum prt_channel_direction_e direction, int slot);
struct prt_packet_s *prt_vdp_packet_new(
    prt_vdp_t *vdp, size_t size, void *data);
struct prt_packet_s *prt_vdp_packet_new_host_to_device(
    prt_vdp_t *vdp, size_t size, void *data);
void prt_vdp_packet_release(prt_vdp_t *vdp, struct prt_packet_s *packet);
void prt_vdp_channel_push(
    prt_vdp_t *vdp, int channel_num, struct prt_packet_s *packet);
struct prt_packet_s *prt_vdp_channel_pop(prt_vdp_t *vdp, int channel_num);
void prt_vdp_channel_off(prt_vdp_t *vdp, int channel_num);
void prt_vdp_channel_on(prt_vdp_t *vdp, int channel_num);
int prt_vdp_ready(prt_vdp_t *vdp);

#ifdef __cplusplus
}
#endif

#endif /* PRT_VDP_H */
