/**
 * @file prt_vsa.h
 * @brief Virtual Systolic Array.
 * @author Jakub Kurzak
 *
 * PULSAR Runtime
 * http://icl.utk.edu/pulsar/
 * Copyright (C) 2012-2015 University of Tennessee.
 */

#ifndef PRT_VSA_H
#define PRT_VSA_H

#include "prt.h"

#ifdef __cplusplus
extern "C" {
#endif

struct prt_vdp_s;
struct prt_proxy_s;
struct prt_config_s;
struct prt_thread_s;
struct prt_device_s;
struct gpu_malloc_s;
struct prt_mapping_s;
enum prt_config_param_e;
enum prt_config_value_e;
enum prt_location_e;

/**
 * @brief The maximum number of VDPs per node.
 * The size of the VSA's hash table of VDPs.
 * Should be a prime number.
 */
#define PRT_VSA_MAX_VDPS_PER_NODE 10003

/**
 * @brief The size of segments allocated by the GPU memory allocator.
 * Setting the unit size to 128 KB, which is 64 x 256 x sizeof(double).
 */
#define PRT_VSA_GPU_ALLOC_UNIT_SIZE 131072

/**
 * @brief The function pointer for the VDP mapping function.
 */
typedef struct prt_mapping_s (*prt_vdp_mapping_t)(int*, void*, int, int);

/**
 * @brief Virtual Systolic Array (VSA)
 * VSA contains global informationa about the system,
 * a local communication proxy, an array of local worker threads,
 * and an array of local accelerator devices.
 */
typedef struct prt_vsa_s {
    int node_rank;                  // Node ID.
    int num_nodes;                  // Total number of nodes.
    int num_threads;                // Number of local threads.
    int num_cores;                  // Total number of worker cores.
    int concurrency;                // Threads+1 if proxy is present.
    pthread_attr_t thread_attr;     // Attributes of worker threads.
    struct prt_thread_s **thread;   // Array of worker threads.
    pthread_barrier_t barrier;      // Thread barrier for synchronization.
    void *global_store;             // VSA's global store accessible by VDPs.
    prt_vdp_mapping_t vdp_mapping;  // VDPs' thread and device mapping function.
    icl_hash_t *vdps_hash;          // VDPs lookup by tuple.
    struct prt_config_s *config;    // VSA's configuration parameters.
    struct prt_proxy_s *proxy;      // VSA's communication proxy.
    icl_list_t **channel_lists;     // List of channels to each node.
    void (*thread_warmup_func)();   // Thread warmup function.
    int num_devices;                // Number of local devices.
    int num_accelerators;           // Total number of accelerators.
    struct prt_device_s **device;   // Array of accelerator devices.
    void (*device_warmup_func)();   // Device warmup function.
    struct gpu_malloc_s **devmem;   // Array of device memory allocators.
} prt_vsa_t;

prt_vsa_t* prt_vsa_new(
    int num_threads, int num_devices, void *global_store,
    struct prt_mapping_s (*vdp_mapping)(int*, void*, int, int));
void prt_vsa_delete(prt_vsa_t *vsa);
void prt_vsa_vdp_insert(prt_vsa_t *vsa, struct prt_vdp_s *vdp);
void prt_vsa_vdp_merge_channels(prt_vsa_t *vsa, struct prt_vdp_s *vdp);
void prt_vsa_vdp_track_tags(prt_vsa_t *vsa, struct prt_vdp_s *vdp);
void prt_vsa_channel_tags(prt_vsa_t *vsa);
void prt_vsa_channel_streams(prt_vsa_t *vsa);
double prt_vsa_run(prt_vsa_t *vsa);
void prt_vsa_config_set(
    prt_vsa_t *vsa, enum prt_config_param_e param, enum prt_config_value_e value);
void prt_vsa_thread_warmup_func_set(prt_vsa_t *vsa, void (*func)());
void prt_vsa_device_warmup_func_set(prt_vsa_t *vsa, void (*func)());
void prt_vsa_devices_warmup(prt_vsa_t *vsa);

#ifdef __cplusplus
}
#endif

#endif /* PRT_VSA_H */
