/**
 * @file svg_trace.c
 * @brief SVG tracing.
 * @author Jakub Kurzak
 *
 * PULSAR Runtime
 * http://icl.utk.edu/pulsar/
 * Copyright (C) 2012-2015 University of Tennessee.
 */

#include "svg_trace.h"

// Per-core events.
static int    eventNumCore  [SVG_TRACE_MAX_CORES];
static double eventStartCore[SVG_TRACE_MAX_CORES][SVG_TRACE_MAX_EVENTS];
static double eventStopCore [SVG_TRACE_MAX_CORES][SVG_TRACE_MAX_EVENTS];
static int    eventColorCore[SVG_TRACE_MAX_CORES][SVG_TRACE_MAX_EVENTS];

// Per-device events.
static int         eventNumDevice  [SVG_TRACE_MAX_DEVICES];
static cudaEvent_t evStartDevice   [SVG_TRACE_MAX_DEVICES][SVG_TRACE_MAX_EVENTS];
static cudaEvent_t evStopDevice    [SVG_TRACE_MAX_DEVICES][SVG_TRACE_MAX_EVENTS];
static double      eventStartDevice[SVG_TRACE_MAX_DEVICES][SVG_TRACE_MAX_EVENTS];
static double      eventStopDevice [SVG_TRACE_MAX_DEVICES][SVG_TRACE_MAX_EVENTS];
static int         eventColorDevice[SVG_TRACE_MAX_DEVICES][SVG_TRACE_MAX_EVENTS];

// Per-device DMA events.
static int         eventNumDMA  [SVG_TRACE_MAX_DEVICES];
static cudaEvent_t evStartDMA   [SVG_TRACE_MAX_DEVICES][SVG_TRACE_MAX_EVENTS];
static cudaEvent_t evStopDMA    [SVG_TRACE_MAX_DEVICES][SVG_TRACE_MAX_EVENTS];
static double      eventStartDMA[SVG_TRACE_MAX_DEVICES][SVG_TRACE_MAX_EVENTS];
static double      eventStopDMA [SVG_TRACE_MAX_DEVICES][SVG_TRACE_MAX_EVENTS];
static int         eventColorDMA[SVG_TRACE_MAX_DEVICES][SVG_TRACE_MAX_EVENTS];

// Memory usage.
static long memoryLevelHost;
static long memoryMaxHost;
static long memoryLevelDevice[SVG_TRACE_MAX_DEVICES];
static long memoryMaxDevice  [SVG_TRACE_MAX_DEVICES];
static pthread_spinlock_t memorySpinlockHost;
static pthread_spinlock_t memorySpinlockDevice[SVG_TRACE_MAX_DEVICES];

static void svg_trace_print_cores(
    int mpi_rank, int mpi_size,
    int num_cores, int num_devices,
    double hscale, double vscale, FILE *trace_file);

static void svg_trace_print_devices(
    int mpi_rank, int mpi_size,
    int num_cores, int num_devices,
    double hscale, double vscale, FILE *trace_file);

static void svg_trace_print_dmas(
    int mpi_rank, int mpi_size,
    int num_cores, int num_devices,
    double hscale, double vscale, FILE *trace_file);

static void svg_trace_print_memory(int mpi_rank, int mpi_size, int num_devices);

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Returns current time.
 *
 * @returns Current Unix time in seconds as a double-precision number.
 */
double get_time_of_day()
{
    struct timeval  time_val;
    struct timezone time_zone;
    gettimeofday(&time_val, &time_zone);
    return (double)(time_val.tv_sec) + (double)(time_val.tv_usec) / 1000000.0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Initializes tracing.
 *
 * @param num_cores -- The number of cores.
 * @param num_devices -- The number of devices.
 */
void svg_trace_init(int num_cores, int num_devices)
{
    assert(num_cores <= SVG_TRACE_MAX_CORES);
    assert(num_devices <= SVG_TRACE_MAX_DEVICES);
    assert(__builtin_popcount(SVG_TRACE_MAX_EVENTS) == 0x01);
    assert(__builtin_popcount(SVG_TRACE_MAX_MEM_SNAPSHOTS) == 0x01);

    // Initialize host memory level spinlock.
    int retval = pthread_spin_init(
        &memorySpinlockHost, PTHREAD_PROCESS_PRIVATE);
    assert(retval == 0);

    int device;
    // FOR each device.
    for (device = 0; device < num_devices; device++) {
        // Set device.
        cudaError_t error = cudaSetDevice(device);
        assert(error == cudaSuccess);

        int event;
        // Initialize device and DMA events.
        for (event = 0; event < SVG_TRACE_MAX_EVENTS; event++) {
            cudaError_t error1 = cudaEventCreate(&evStartDevice[device][event]);
            cudaError_t error2 = cudaEventCreate(&evStopDevice [device][event]);
            assert(error1 == cudaSuccess);
            assert(error2 == cudaSuccess);
            error1 = cudaEventCreate(&evStartDMA[device][event]);
            error2 = cudaEventCreate(&evStopDMA [device][event]);
            assert(error1 == cudaSuccess);
            assert(error2 == cudaSuccess);
        }
        // Initialize device memory level spinlock.
        int retval = pthread_spin_init(
            &memorySpinlockDevice[device], PTHREAD_PROCESS_PRIVATE);
        assert(retval == 0);
    }
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Starts tracing of a CPU event.
 *
 * @param thread_rank -- The rank of the thread.
 */
void svg_trace_start_cpu(int thread_rank)
{
    assert(thread_rank < SVG_TRACE_MAX_CORES);
    eventStartCore[thread_rank][eventNumCore[thread_rank]] = get_time_of_day();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Stops tracing a CPU event.
 *
 * @param thread_rank -- The rank of the thread.
 * @param color -- The RGB color of the SVG box.
 */
void svg_trace_stop_cpu(int thread_rank, int color)
{
    assert(thread_rank < SVG_TRACE_MAX_CORES);
    eventStopCore [thread_rank][eventNumCore[thread_rank]] = get_time_of_day();
    eventColorCore[thread_rank][eventNumCore[thread_rank]] = color;
    eventNumCore[thread_rank]++;
    eventNumCore[thread_rank] &= (SVG_TRACE_MAX_EVENTS-1);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Starts tracing a GPU event.
 *
 * @param stream -- The stream of the event.
 */
void svg_trace_start_gpu(cudaStream_t stream)
{
    int device;
    cudaGetDevice(&device);
    assert(device < SVG_TRACE_MAX_DEVICES);
    cudaError_t error = cudaEventRecord(
        evStartDevice[device][eventNumDevice[device]], stream);
    assert(error == cudaSuccess);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Stops tracing a GPU event.
 *
 * @param stream -- The stream of the event.
 * @param color -- The RGB color of the SVG box.
 */
void svg_trace_stop_gpu(cudaStream_t stream, int color)
{
    int device;
    cudaGetDevice(&device);
    assert(device < SVG_TRACE_MAX_DEVICES);
    cudaError_t error = cudaEventRecord(
        evStopDevice[device][eventNumDevice[device]], stream);
    assert(error == cudaSuccess);
    eventColorDevice[device][eventNumDevice[device]] = color;
    eventNumDevice[device]++;
    eventNumDevice[device] &= (SVG_TRACE_MAX_EVENTS-1);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Starts tracing a DMA event.
 *
 * @param stream -- The stream of the event.
 */
void svg_trace_start_dma(cudaStream_t stream)
{
    int device;
    cudaGetDevice(&device);
    assert(device < SVG_TRACE_MAX_DEVICES);
    cudaError_t error = cudaEventRecord(
        evStartDMA[device][eventNumDMA[device]], stream);
    assert(error == cudaSuccess);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Stops tracing a DMA event.
 *
 * @param stream -- The stream of the event.
 * @param color -- The RGB color of the SVG box.
 */
void svg_trace_stop_dma(cudaStream_t stream, int color)
{
    int device;
    cudaGetDevice(&device);
    assert(device < SVG_TRACE_MAX_DEVICES);
    cudaError_t error = cudaEventRecord(
        evStopDMA[device][eventNumDMA[device]], stream);
    assert(error == cudaSuccess);
    eventColorDMA[device][eventNumDMA[device]] = color;
    eventNumDMA[device]++;
    eventNumDMA[device] &= (SVG_TRACE_MAX_EVENTS-1);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Registers host memory usage.
 * The operation has to be atomic, because it can be invoked by a callback.
 * Because there are two variables to keep track of, the level and the maximum,
 * doing it with atomics is not worth it. Using a spinlock instead.
 *
 * @param delta -- The change of host memory usage in bytes.
 */
void svg_trace_memory_host(long delta)
{
    pthread_spin_lock(&memorySpinlockHost);
    memoryLevelHost += delta;
    if (memoryLevelHost > memoryMaxHost)
        memoryMaxHost = memoryLevelHost;
    pthread_spin_unlock(&memorySpinlockHost);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Register device memory usage.
 *
 * @param delta -- The change of host memory usage in bytes.
 */
void svg_trace_memory_device(long delta)
{
    int device;
    cudaGetDevice(&device);
    assert(device < SVG_TRACE_MAX_DEVICES);
    pthread_spin_lock(&memorySpinlockDevice[device]);
    memoryLevelDevice[device] += delta;
    if (memoryLevelDevice[device] > memoryMaxDevice[device])
        memoryMaxDevice[device] = memoryLevelDevice[device];
    pthread_spin_unlock(&memorySpinlockDevice[device]);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Finishes tracing.
 * Collects traces from all nodes.
 * Writes the combined trace to an SVG file.
 */
void svg_trace_finish(int num_cores, int num_devices)
{
    int mpi_rank;
    int mpi_size;
    MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);
    MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);

    int proc;
    int core;
    int event;
    int device;
    int snapshot;

    double max_time_in = 0.0;
    // Find maximum timestamp.
    // Compute horizontal scaling factor.
    for (core = 0; core < SVG_TRACE_MAX_CORES; core++) {
        double time =
            eventStopCore [core][eventNumCore[core]-1] -
            eventStopCore[0][0];
        if (time > max_time_in)
            max_time_in = time;
    }
    double max_time;
#ifdef MPI
    MPI_Reduce(
        &max_time_in, &max_time, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
#else
    max_time = max_time_in;
#endif
    double hscale_cpu = 2000.0 / max_time;
    double hscale_gpu = hscale_cpu / 1000.0;

    // Compute the vertical scaling factor.
    double vscale = 1000.0 / ((num_cores+num_devices*2)*mpi_size);

    FILE *trace_file;
    if (mpi_rank == 0) {
        char trace_file_name[SVG_TRACE_FILE_NAME_SIZE];
        sprintf(trace_file_name, "trace_%d.svg", (int)(time(NULL)));
        trace_file = fopen(trace_file_name, "w");
        assert(trace_file != NULL);
        fprintf(trace_file,
            "<svg width=\"200mm\" height=\"100mm\" viewBox=\"0 0 2000 1000\">\n"
            "  <g>\n");
    }
    // Print cores' traces.
    svg_trace_print_cores(
        mpi_rank, mpi_size,
        num_cores, num_devices,
        hscale_cpu, vscale, trace_file);

    // Print devices' traces.
    svg_trace_print_devices(
        mpi_rank, mpi_size,
        num_cores, num_devices,
        hscale_gpu, vscale, trace_file);

    // Print DMAs' traces.
    svg_trace_print_dmas(
        mpi_rank, mpi_size,
        num_cores, num_devices,
        hscale_gpu, vscale, trace_file);

    if (mpi_rank == 0) {
        fprintf(trace_file,
            "  </g>\n"
            "</svg>\n");
        fclose(trace_file);
    }
    // Print memory usage info.
    svg_trace_print_memory(mpi_rank, mpi_size, num_devices);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Writes cores' traces.
 *
 * @param mpi_rank -- The MPI rank of the node.
 * @param mpi_size -- The MPI size of the run.
 * @param num_cores -- The number of cores.
 * @param num_devices -- The number of devices.
 * @param hscale -- The horizontal scaling factor.
 * @param vscale -- The vertical scaling factor.
 * @param trace_file -- The trace file handle.
 */
static void svg_trace_print_cores(
    int mpi_rank, int mpi_size,
    int num_cores, int num_devices,
    double hscale, double vscale, FILE *trace_file)
{
    int proc;
    int core;
    int event;
    if (mpi_rank == 0) {
      for (proc = 0; proc < mpi_size; proc++) {
        if (proc > 0) {
            // Receive events.
            MPI_Recv(
                &eventNumCore[0],
                SVG_TRACE_MAX_CORES, MPI_INT,
                proc, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            MPI_Recv(
                &eventStartCore[0][0],
                SVG_TRACE_MAX_CORES*SVG_TRACE_MAX_EVENTS, MPI_DOUBLE,
                proc, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            MPI_Recv(
                &eventStopCore[0][0],
                SVG_TRACE_MAX_CORES*SVG_TRACE_MAX_EVENTS, MPI_DOUBLE,
                proc, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            MPI_Recv(
                &eventColorCore[0][0],
                SVG_TRACE_MAX_CORES*SVG_TRACE_MAX_EVENTS, MPI_INT,
                proc, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        }
        for (core = 0; core < num_cores; core++) {
          for (event = 0; event < eventNumCore[core]; event++) {
            double start = eventStartCore[core][event] - eventStopCore[0][0];
            double stop =  eventStopCore [core][event] - eventStopCore[0][0];
            double width = (stop-start) * hscale;
            int color = eventColorCore[core][event];
            int thread = proc*(num_cores+2*num_devices);
            thread += core;
            fprintf(trace_file,
                "    <rect "
                "x=\"%.2lf\" y=\"%.0lf\" width=\"%.2lf\" height=\"%.0lf\" "
                "fill=\"#%06x\" stroke=\"#%06x\" stroke-width=\"0.5\"/>\n",
                start * hscale,
                thread * vscale,
                width < 2.0 ? 2.0 : width,
                vscale * 0.9,
                abs(color),
                color < 0 ? color : 0);
          }
        }
      }
    }
    else {
        // Send events.
        MPI_Send(
            &eventNumCore[0],
            SVG_TRACE_MAX_CORES,
            MPI_INT, 0, 0, MPI_COMM_WORLD);
        MPI_Send(
            &eventStartCore[0][0],
            SVG_TRACE_MAX_CORES*SVG_TRACE_MAX_EVENTS,
            MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
        MPI_Send(
            &eventStopCore[0][0],
            SVG_TRACE_MAX_CORES*SVG_TRACE_MAX_EVENTS,
            MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
        MPI_Send(
            &eventColorCore[0][0],
            SVG_TRACE_MAX_CORES*SVG_TRACE_MAX_EVENTS,
            MPI_INT, 0, 0, MPI_COMM_WORLD);
    }
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Writes devices' traces.
 *
 * @param mpi_rank -- The MPI rank of the node.
 * @param mpi_size -- The MPI size of the run.
 * @param num_cores -- The number of cores.
 * @param num_devices -- The number of devices.
 * @param hscale -- The horizontal scaling factor.
 * @param vscale -- The vertical scaling factor.
 * @param trace_file -- The trace file handle.
 */
static void svg_trace_print_devices(
    int mpi_rank, int mpi_size,
    int num_cores, int num_devices,
    double hscale, double vscale, FILE *trace_file)
{
    int event;
    int device;
    // Convert cudaEvent_t to double.
    for (device = 0; device < num_devices; device++) {
        cudaError_t error = cudaSetDevice(device);
        assert(error == cudaSuccess);
        for (event = 0; event < eventNumDevice[device]; event++)
        {
            float fstart;
            float fstop;
            cudaEventElapsedTime(&fstart,
                evStartDevice[device][0], evStartDevice[device][event]);
            cudaEventElapsedTime(&fstop,
                evStartDevice[device][0], evStopDevice[device][event]);
            double dstart = (double)fstart;
            double dstop  = (double)fstop;
            eventStartDevice[device][event] = dstart;
            eventStopDevice[device][event] = dstop;
        }
        eventStartDevice[device][0] = 0.0;
        eventStopDevice[device][0] = 0.0;
    }
    int proc;
    if (mpi_rank == 0) {
      for (proc = 0; proc < mpi_size; proc++) {
        if (proc > 0) {
            // Receive events.
            MPI_Recv(
                &eventNumDevice[0],
                SVG_TRACE_MAX_DEVICES, MPI_INT,
                proc, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            MPI_Recv(
                &eventStartDevice[0][0],
                SVG_TRACE_MAX_DEVICES*SVG_TRACE_MAX_EVENTS, MPI_DOUBLE,
                proc, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            MPI_Recv(
                &eventStopDevice[0][0],
                SVG_TRACE_MAX_DEVICES*SVG_TRACE_MAX_EVENTS, MPI_DOUBLE,
                proc, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            MPI_Recv(
                &eventColorDevice[0][0],
                SVG_TRACE_MAX_DEVICES*SVG_TRACE_MAX_EVENTS, MPI_INT,
                proc, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        }
        for (device = 0; device < num_devices; device++) {
          for (event = 1; event < eventNumDevice[device]; event++) {
            double start = eventStartDevice[device][event];
            double stop =  eventStopDevice [device][event];
            double width = (stop-start) * hscale;
            int color = eventColorDevice[device][event];
            int accel = proc*(num_cores+2*num_devices);
            accel += num_cores+device;
            fprintf(trace_file,
                "    <rect "
                "x=\"%.2lf\" y=\"%.0lf\" width=\"%.2lf\" height=\"%.0lf\" "
                "fill=\"#%06x\" stroke=\"#%06x\" stroke-width=\"0.5\"/>\n",
                start * hscale,
                accel * vscale,
                width < 2.0 ? 2.0 : width,
                vscale * 0.9,
                abs(color),
                color < 0 ? color : 0);
          }
        }
      }
    }
    else {
        // Send events.
        MPI_Send(
            &eventNumDevice[0],
            SVG_TRACE_MAX_DEVICES,
            MPI_INT, 0, 0, MPI_COMM_WORLD);
        MPI_Send(
            &eventStartDevice[0][0],
            SVG_TRACE_MAX_DEVICES*SVG_TRACE_MAX_EVENTS,
            MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
        MPI_Send(
            &eventStopDevice[0][0],
            SVG_TRACE_MAX_DEVICES*SVG_TRACE_MAX_EVENTS,
            MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
        MPI_Send(
            &eventColorDevice[0][0],
            SVG_TRACE_MAX_DEVICES*SVG_TRACE_MAX_EVENTS,
            MPI_INT, 0, 0, MPI_COMM_WORLD);
    }
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Writes DMAs' traces.
 *
 * @param mpi_rank -- The MPI rank of the node.
 * @param mpi_size -- The MPI size of the run.
 * @param num_cores -- The number of cores.
 * @param num_devices -- The number of devices.
 * @param hscale -- The horizontal scaling factor.
 * @param vscale -- The vertical scaling factor.
 * @param trace_file -- The trace file handle.
 */
static void svg_trace_print_dmas(
    int mpi_rank, int mpi_size,
    int num_cores, int num_devices,
    double hscale, double vscale, FILE *trace_file)
{
    int event;
    int device;
    // Convert cudaEvent_t to double.
    for (device = 0; device < num_devices; device++) {
        cudaError_t error = cudaSetDevice(device);
        assert(error == cudaSuccess);
        for (event = 0; event < eventNumDMA[device]; event++)
        {
            float fstart;
            float fstop;
            cudaEventElapsedTime(&fstart,
                evStartDMA[device][0], evStartDMA[device][event]);
            cudaEventElapsedTime(&fstop,
                evStartDMA[device][0], evStopDMA[device][event]);
            double dstart = (double)fstart;
            double dstop  = (double)fstop;
            eventStartDMA[device][event] = dstart;
            eventStopDMA[device][event] = dstop;
        }
        eventStartDMA[device][0] = 0.0;
        eventStopDMA[device][0] = 0.0;
    }
    int proc;
    if (mpi_rank == 0) {
      for (proc = 0; proc < mpi_size; proc++) {
        if (proc > 0) {
            // Receive events.
            MPI_Recv(
                &eventNumDMA[0],
                SVG_TRACE_MAX_DEVICES, MPI_INT,
                proc, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            MPI_Recv(
                &eventStartDMA[0][0],
                SVG_TRACE_MAX_DEVICES*SVG_TRACE_MAX_EVENTS, MPI_DOUBLE,
                proc, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            MPI_Recv(
                &eventStopDMA[0][0],
                SVG_TRACE_MAX_DEVICES*SVG_TRACE_MAX_EVENTS, MPI_DOUBLE,
                proc, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            MPI_Recv(
                &eventColorDMA[0][0],
                SVG_TRACE_MAX_DEVICES*SVG_TRACE_MAX_EVENTS, MPI_INT,
                proc, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        }
        for (device = 0; device < num_devices; device++) {
          for (event = 1; event < eventNumDMA[device]; event++) {
            double start = eventStartDMA[device][event];
            double stop =  eventStopDMA [device][event];
            double width = (stop-start) * hscale;
            int color = eventColorDMA[device][event];
            int accel = proc*(num_cores+2*num_devices);
            accel += num_cores+num_devices+device;
            fprintf(trace_file,
                "    <rect "
                "x=\"%.2lf\" y=\"%.0lf\" width=\"%.2lf\" height=\"%.0lf\" "
                "fill=\"#%06x\" stroke=\"#%06x\" stroke-width=\"0.5\"/>\n",
                start * hscale,
                accel * vscale,
                width < 2.0 ? 2.0 : width,
                vscale * 0.9,
                abs(color),
                color < 0 ? color : 0);
          }
        }
      }
    }
    else {
        // Send events.
        MPI_Send(
            &eventNumDMA[0],
            SVG_TRACE_MAX_DEVICES,
            MPI_INT, 0, 0, MPI_COMM_WORLD);
        MPI_Send(
            &eventStartDMA[0][0],
            SVG_TRACE_MAX_DEVICES*SVG_TRACE_MAX_EVENTS,
            MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
        MPI_Send(
            &eventStopDMA[0][0],
            SVG_TRACE_MAX_DEVICES*SVG_TRACE_MAX_EVENTS,
            MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
        MPI_Send(
            &eventColorDMA[0][0],
            SVG_TRACE_MAX_DEVICES*SVG_TRACE_MAX_EVENTS,
            MPI_INT, 0, 0, MPI_COMM_WORLD);
    }
}

////////////////////////////////////////////////////////////////////////////////
/**
 * @brief Writes memory usage.
 *
 * @param mpi_rank -- The MPI rank of the node.
 * @param mpi_size -- The MPI size of the run.
 * @param num_devices -- The number of devices.
 */
static void svg_trace_print_memory(int mpi_rank, int mpi_size, int num_devices)
{
    int rank;
    printf("\n");
    for (rank = 0; rank < mpi_size; rank++) {
        if (rank == mpi_rank) {

            // Print maximum host memory level.
            if (labs(memoryMaxHost) < 1024)
                printf("Host %d max:\t%ld B\n",
                    rank, memoryMaxHost);
            else if (labs(memoryMaxHost) < 1024*1024)
                printf("Host %d max:\t%.1lf KB\n",
                    rank, memoryMaxHost/1024.0);
            else if (labs(memoryMaxHost) < 1024*1024*1024)
                printf("Host %d max:\t%.1lf MB\n",
                    rank, memoryMaxHost/1024.0/1024.0);
            else
                printf("Host %d max:\t%.1lf GB\n",
                    rank, memoryMaxHost/1024.0/1024.0/1024.0);

            // Print final host memory level.
            if (labs(memoryLevelHost) < 1024)
                printf("Host %d end:\t%ld B\n",
                    rank, memoryLevelHost);
            else if (labs(memoryLevelHost) < 1024*1024)
                printf("Host %d end:\t%.1lf KB\n",
                    rank, memoryLevelHost/1024.0);
            else if (labs(memoryLevelHost) < 1024*1024*1024)
                printf("Host %d end:\t%.1lf MB\n",
                    rank, memoryLevelHost/1024.0/1024.0);
            else
                printf("Host %d end:\t%.1lf GB\n",
                    rank, memoryLevelHost/1024.0/1024.0/1024.0);

            int device;
            for (device = 0; device < num_devices; device++) {

                // Print maximum host memory level.
                if (labs(memoryMaxDevice[device]) < 1024)
                    printf("  Device %d max:\t%ld B\n",
                        device, memoryMaxDevice[device]);
                else if (labs(memoryMaxDevice[device]) < 1024*1024)
                    printf("  Device %d max:\t%.1lf KB\n",
                        device, memoryMaxDevice[device]/1024.0);
                else if (labs(memoryMaxDevice[device]) < 1024*1024*1024)
                    printf("  Device %d max:\t%.1lf MB\n",
                        device, memoryMaxDevice[device]/1024.0/1024.0);
                else
                    printf("  Device %d max:\t%.1lf GB\n",
                        device, memoryMaxDevice[device]/1024.0/1024.0/1024.0);

                // Print final host memory level.
                if (labs(memoryLevelDevice[device]) < 1024)
                    printf("  Device %d end:\t%ld B\n",
                        device, memoryLevelDevice[device]);
                else if (labs(memoryLevelDevice[device]) < 1024*1024)
                    printf("  Device %d end:\t%.1lf KB\n",
                        device, memoryLevelDevice[device]/1024.0);
                else if (labs(memoryLevelDevice[device]) < 1024*1024*1024)
                    printf("  Device %d end:\t%.1lf MB\n",
                        device, memoryLevelDevice[device]/1024.0/1024.0);
                else
                    printf("  Device %d end:\t%.1lf GB\n",
                        device, memoryLevelDevice[device]/1024.0/1024.0/1024.0);

            }
            MPI_Barrier(MPI_COMM_WORLD);
        }
    }
}
